from __future__ import annotations

from collections.abc import (
    ItemsView,
    Iterable,
    KeysView,
    Mapping,
    MutableMapping,
    ValuesView,
)
from typing import (
    Generic,
    TypeVar,
)

KT = TypeVar("KT")
VT = TypeVar("VT")


class bidict(Generic[KT, VT]):
    """
    Bidirectional Dictionary
    All keys and values must be unique (bijective one to one mapping).

    To go from key to value use `bidict.fwd`.
    To go from value to key use `bidict.bwd`.
    """

    def __init__(self, seq: Mapping[KT, VT] = None, /, *, fwd=None, bwd=None):
        if fwd is not None and bwd is not None:
            self.__fwd = fwd
            self.__bwd = bwd
        else:
            self.__fwd = dict(seq) if seq is not None else dict()
            self.__bwd = {v: k for k, v in self.__fwd.items()}

        if len(self.__fwd) != len(self.__bwd) != len(seq):
            raise ValueError(
                "Input sequence contains duplicate key value pairs. Mapping must be"
                " unique."
            )

        self.fwd = _BidictInterface(
            self.__fwd, self.__bwd
        )  # type: _BidictInterface[KT, VT]
        self.bwd = _BidictInterface(
            self.__bwd, self.__fwd
        )  # type: _BidictInterface[VT, KT]

    def __copy__(self):
        return bidict(fwd=self.__fwd.copy(), bwd=self.__bwd.copy())

    def __len__(self):
        return len(self.__fwd)

    def clear(self):
        self.__fwd.clear()
        self.__bwd.clear()


class _BidictInterface(MutableMapping[KT, VT]):
    def __init__(self, fwd: dict[KT, VT], bwd: dict[VT, KT]):
        self.__fwd = fwd
        self.__bwd = bwd

    def __setitem__(self, key: KT, value: VT):
        if key in self.__fwd:
            fwd_value = self.__fwd[key]
            del self.__bwd[fwd_value]
        if value in self.__bwd:
            raise ValueError(f"Duplicate value '{value}'. Mapping must be unique.")
        self.__fwd[key] = value
        self.__bwd[value] = key

    def __getitem__(self, key: KT) -> VT:
        return self.__fwd[key]

    def __delitem__(self, key: KT):
        value = self.__fwd[key]
        del self.__fwd[key]
        del self.__bwd[value]

    def __iter__(self) -> Iterable[KT]:
        yield from self.__fwd.__iter__()

    def __len__(self) -> int:
        return len(self.__fwd)

    def __contains__(self, item) -> bool:
        return item in self.__fwd

    def items(self) -> ItemsView[KT, VT]:
        return self.__fwd.items()

    def keys(self) -> KeysView[KT]:
        return self.__fwd.keys()

    def values(self) -> ValuesView[VT]:
        return self.__fwd.values()



from __future__ import annotations

import copy
from functools import partial, reduce, wraps
from typing import Any

from pydiverse.transform.core.expressions import (
    Column,
    LambdaColumn,
    unwrap_symbolic_expressions,
)
from pydiverse.transform.core.util import bidict, traverse


class Pipeable:
    def __init__(self, f=None, calls=None):
        if f is not None:
            if calls is not None:
                raise ValueError
            self.calls = [f]
        else:
            self.calls = calls

    def __rshift__(self, other) -> Pipeable:
        """
        Pipeable >> other
        -> Lazy. Extend pipe.
        """
        if isinstance(other, Pipeable):
            return Pipeable(calls=self.calls + other.calls)
        elif callable(other):
            return Pipeable(calls=self.calls + [other])

        raise RuntimeError

    def __rrshift__(self, other):
        """
        other >> Pipeable
        -> Eager.
        """
        if callable(other):
            return Pipeable(calls=[other] + self.calls)
        return self(other)

    def __call__(self, arg):
        return reduce(lambda x, f: f(x), self.calls, arg)


class inverse_partial(partial):
    """
    Just like partial, but the arguments get applied to the back instead of the front.
    This means that a function `def x(a, b, c)` decorated with `@inverse_partial(1, 2)`
    that gets called with `x(0)` is equivalent to calling `x(0, 1, 2)` on the non
    decorated function.
    """

    def __call__(self, /, *args, **keywords):
        keywords = {**self.keywords, **keywords}
        #                   ↙ *args moved to front.
        return self.func(*args, *self.args, **keywords)


def verb(func):
    from pydiverse.transform.core.table import Table

    def copy_tables(arg: Any = None):
        return traverse(arg, lambda x: copy.copy(x) if isinstance(x, Table) else x)

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Copy Table objects to prevent mutating them
        # This can be the case if the user uses __setitem__ inside the verb
        def f(*args, **kwargs):
            args = copy_tables(args)
            kwargs = copy_tables(kwargs)
            return func(*args, **kwargs)

        f = inverse_partial(f, *args, **kwargs)  # Bind arguments
        return Pipeable(f)

    return wrapper


def builtin_verb(backends=None):
    def wrap_and_unwrap(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            args = list(args)
            args = unwrap_symbolic_expressions(args)
            if len(args):
                args[0] = col_to_table(args[0])
            args = unwrap_tables(args)

            kwargs = unwrap_symbolic_expressions(kwargs)
            kwargs = unwrap_tables(kwargs)

            return wrap_tables(func(*args, **kwargs))

        return wrapper

    def check_backend(func):
        if backends is None:
            return func

        @wraps(func)
        def wrapper(*args, **kwargs):
            assert len(args) > 0
            impl = args[0]._impl
            if isinstance(impl, backends):
                return func(*args, **kwargs)
            raise TypeError(f"Backend {impl} not supported for verb '{func.__name__}'.")

        return wrapper

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            f = func
            f = wrap_and_unwrap(f)  # Convert from Table to Impl and back
            f = check_backend(f)  # Check type of backend
            f = inverse_partial(f, *args, **kwargs)  # Bind arguments
            return Pipeable(f)  # Make pipeable

        return wrapper

    return decorator


# Helper


def col_to_table(arg: Any = None):
    """
    Takes a single argument and if it is a column, replaces it with a table
    implementation that only contains that one column.

    This allows for more eager style code where you perform operations on
    columns like with the following example::

        def get_c(b, tB):
            tC = b >> left_join(tB, b == tB.b)
            return tC[tB.c]
        feature_col = get_c(tblA.b, tblB)

    """
    from pydiverse.transform.core.verbs import select

    if isinstance(arg, Column):
        tbl = (arg.table >> select(arg))._impl
        col = tbl.get_col(arg)

        tbl.available_cols = {col.uuid}
        tbl.named_cols = bidict({col.name: col.uuid})
        return tbl
    elif isinstance(arg, LambdaColumn):
        raise ValueError("Can't start a pipe with a lambda column.")

    return arg


def unwrap_tables(arg: Any = None):
    """
    Takes an instance or collection of `Table` objects and replaces them with
    their implementation.
    """
    from pydiverse.transform.core.table import Table

    return traverse(arg, lambda x: x._impl if isinstance(x, Table) else x)


def wrap_tables(arg: Any = None):
    """
    Takes an instance or collection of `AbstractTableImpl` objects and wraps
    them in a `Table` object. This is an inverse to the `unwrap_tables` function.
    """
    from pydiverse.transform.core.table import Table
    from pydiverse.transform.core.table_impl import AbstractTableImpl

    return traverse(arg, lambda x: Table(x) if isinstance(x, AbstractTableImpl) else x)



from __future__ import annotations

from abc import ABC, abstractmethod

from pydiverse.transform._typing import T
from pydiverse.transform.errors import ExpressionTypeError


class DType(ABC):
    def __init__(self, *, const: bool = False, vararg: bool = False):
        self.const = const
        self.vararg = vararg

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if self.const != other.const:
            return False
        if self.vararg != other.vararg:
            return False
        if self.name != other.name:
            return False
        return True

    def __hash__(self):
        return hash((self.name, self.const, self.vararg, type(self).__qualname__))

    def __repr__(self):
        dtype_str = ""
        if self.const:
            dtype_str += "const "
        dtype_str += self.name
        if self.vararg:
            dtype_str += "..."

        return dtype_str

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    def without_modifiers(self: T) -> T:
        """Returns a copy of `self` with all modifiers removed"""
        return type(self)()

    def same_kind(self, other: DType) -> bool:
        """Check if `other` is of the same type as self.

        More specifically, `other` must be a stricter subtype of `self`.
        """
        if not isinstance(other, type(self)):
            return False

        # self.const -> other.const
        if self.const and not other.const:
            return False

        # other.vararg -> self.vararg
        if other.vararg and not self.vararg:
            return False

        return True

    def can_promote_to(self, other: DType) -> bool:
        return other.same_kind(self)


class Int(DType):
    name = "int"

    def can_promote_to(self, other: DType) -> bool:
        if super().can_promote_to(other):
            return True

        # int can be promoted to float
        if Float().same_kind(other):
            if other.const and not self.const:
                return False

            return True

        return False


class Float(DType):
    name = "float"


class String(DType):
    name = "str"


class Bool(DType):
    name = "bool"


class DateTime(DType):
    name = "datetime"


class Template(DType):
    name = None

    def __init__(self, name, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def without_modifiers(self: T) -> T:
        return type(self)(self.name)

    def same_kind(self, other: DType) -> bool:
        if not super().same_kind(other):
            return False

        return self.name == other.name

    def modifiers_compatible(self, other: DType) -> bool:
        """
        Check if another dtype object is compatible with the modifiers of the template.
        """
        if self.const and not other.const:
            return False
        return True


class NoneDType(DType):
    """DType used to represent the `None` value."""

    name = "none"


def dtype_from_string(t: str) -> DType:
    parts = [part for part in t.split(" ") if part]

    is_const = False
    is_vararg = False

    # Handle vararg
    if parts[-1] == "...":
        del parts[-1]
        is_vararg = True
    elif parts[-1].endswith("..."):
        parts[-1] = parts[-1][:-3]
        is_vararg = True

    *modifiers, base_type = parts

    # Handle modifiers
    for modifier in modifiers:
        if modifier == "const":
            is_const = True
        else:
            raise ValueError(f"Unknown type modifier '{modifier}'.")

    # Handle type
    is_template = len(base_type) == 1 and 65 <= ord(base_type) <= 90

    if is_template:
        return Template(base_type, const=is_const, vararg=is_vararg)

    if base_type == "int":
        return Int(const=is_const, vararg=is_vararg)
    if base_type == "float":
        return Float(const=is_const, vararg=is_vararg)
    if base_type == "str":
        return String(const=is_const, vararg=is_vararg)
    if base_type == "bool":
        return Bool(const=is_const, vararg=is_vararg)
    if base_type == "datetime":
        return DateTime(const=is_const, vararg=is_vararg)
    if base_type == "none":
        return NoneDType(const=is_const, vararg=is_vararg)

    raise ValueError(f"Unknown type '{base_type}'")


def promote_dtypes(dtypes: list[DType]) -> DType:
    if len(dtypes) == 0:
        raise ValueError("Expected non empty list of dtypes")

    promoted = dtypes[0]
    for dtype in dtypes[1:]:
        if isinstance(dtype, NoneDType):
            continue
        if isinstance(promoted, NoneDType):
            promoted = dtype
            continue

        if dtype.can_promote_to(promoted):
            continue
        if promoted.can_promote_to(dtype):
            promoted = dtype
            continue

        raise ExpressionTypeError(f"Incompatible types {dtype} and {promoted}.")

    return promoted



from __future__ import annotations

import functools
import inspect
import itertools
import operator as py_operator
import uuid
import warnings
from collections.abc import Iterable
from functools import reduce
from typing import TYPE_CHECKING, Callable

import sqlalchemy as sa
from sqlalchemy import sql

from pydiverse.transform import ops
from pydiverse.transform.core import dtypes
from pydiverse.transform.core.expressions import (
    Column,
    LiteralColumn,
    SymbolicExpression,
    iterate_over_expr,
)
from pydiverse.transform.core.expressions.translator import TypedValue
from pydiverse.transform.core.table_impl import ColumnMetaData
from pydiverse.transform.core.util import OrderingDescriptor, translate_ordering
from pydiverse.transform.errors import AlignmentError, FunctionTypeError
from pydiverse.transform.lazy.lazy_table import JoinDescriptor, LazyTableImpl
from pydiverse.transform.ops import OPType

if TYPE_CHECKING:
    from pydiverse.transform.core.expressions import FunctionCall
    from pydiverse.transform.core.registry import TypedOperatorImpl


class SQLTableImpl(LazyTableImpl):
    """SQL backend

    Attributes:
        tbl: The underlying SQLAlchemy table object.
        engine: The SQLAlchemy engine.
        sql_columns: A dict mapping from uuids to SQLAlchemy column objects
            (only those contained in `tbl`).

        alignment_hash: A hash value that allows checking if two tables are
            'aligned'. In the case of SQL this means that two tables NUST NOT
            share the same alignment hash unless they were derived from the
            same table(s) and are guaranteed to have the same number of columns
            in the same order. In other words: Two tables MUST only have the
            same alignment hash if a literal column derived from one of them
            can be used by the other table and produces the same result.
    """

    __registered_dialects: dict[str, type[SQLTableImpl]] = {}
    _dialect_name: str

    def __new__(cls, *args, **kwargs):
        if cls != SQLTableImpl or (not args and not kwargs):
            return super().__new__(cls)

        signature = inspect.signature(cls.__init__)
        engine = signature.bind(None, *args, **kwargs).arguments["engine"]

        # If calling SQLTableImpl(engine), then we want to dynamically instantiate
        # the correct dialect specific subclass based on the engine dialect.
        if isinstance(engine, str):
            dialect = sa.engine.make_url(engine).get_dialect().name
        else:
            dialect = engine.dialect.name

        dialect_specific_cls = SQLTableImpl.__registered_dialects.get(dialect, cls)
        return super(SQLTableImpl, dialect_specific_cls).__new__(dialect_specific_cls)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Whenever a new subclass if SQLTableImpl is defined, it must contain the
        # `_dialect_name` attribute. This allows us to dynamically instantiate it
        # when calling SQLTableImpl(engine) based on the dialect name found
        # in the engine url (see __new__).
        dialect_name = getattr(cls, "_dialect_name", None)
        if dialect_name is None:
            raise ValueError(
                "All subclasses of SQLTableImpl must have a `_dialect_name` attribute."
                f" But {cls.__name__}._dialect_name is None."
            )

        if dialect_name in SQLTableImpl.__registered_dialects:
            warnings.warn(
                f"Already registered a SQLTableImpl for dialect {dialect_name}"
            )
        SQLTableImpl.__registered_dialects[dialect_name] = cls

    def __init__(
        self,
        engine: sa.Engine | str,
        table,
        _dtype_hints: dict[str, dtypes.DType] = None,
    ):
        self.engine = sa.create_engine(engine) if isinstance(engine, str) else engine
        tbl = self._create_table(table, self.engine)

        columns = {
            col.name: Column(
                name=col.name,
                table=self,
                dtype=self._get_dtype(col, hints=_dtype_hints),
            )
            for col in tbl.columns
        }

        self.replace_tbl(tbl, columns)
        super().__init__(name=self.tbl.name, columns=columns)

    def is_aligned_with(self, col: Column | LiteralColumn) -> bool:
        if isinstance(col, Column):
            if not isinstance(col.table, type(self)):
                return False
            return col.table.alignment_hash == self.alignment_hash

        if isinstance(col, LiteralColumn):
            return all(
                self.is_aligned_with(atom)
                for atom in iterate_over_expr(col.expr, expand_literal_col=True)
                if isinstance(atom, Column)
            )

        raise ValueError

    @classmethod
    def _html_repr_expr(cls, expr):
        if isinstance(expr, sa.sql.ColumnElement):
            return str(expr.compile(compile_kwargs={"literal_binds": True}))
        return super()._html_repr_expr(expr)

    @staticmethod
    def _create_table(tbl, engine=None):
        """Return a sa.Table

        :param tbl: a sa.Table or string of form 'table_name'
            or 'schema_name.table_name'.
        """
        if isinstance(tbl, sa.sql.FromClause):
            return tbl

        if not isinstance(tbl, str):
            raise ValueError(
                "tbl must be a sqlalchemy Table or string, but was %s" % type(tbl)
            )

        schema, table_name = tbl.split(".") if "." in tbl else [None, tbl]
        return sa.Table(
            table_name,
            sa.MetaData(),
            schema=schema,
            autoload_with=engine,
        )

    @staticmethod
    def _get_dtype(
        col: sa.Column, hints: dict[str, dtypes.DType] = None
    ) -> dtypes.DType:
        """Determine the dtype of a column.

        :param col: The sqlalchemy column object.
        :param hints: In some situations sqlalchemy can't determine the dtype of
            a column. Instead of throwing an exception we can use these type
            hints as a fallback.
        :return: Appropriate dtype string.
        """

        type_ = col.type
        if isinstance(type_, sa.Integer):
            return dtypes.Int()
        if isinstance(type_, sa.Numeric):
            return dtypes.Float()
        if isinstance(type_, sa.String):
            return dtypes.String()
        if isinstance(type_, sa.Boolean):
            return dtypes.Bool()
        if isinstance(type_, sa.DateTime):
            return dtypes.DateTime()
        if isinstance(type_, sa.Date):
            raise NotImplementedError("Unsupported type: Date")
        if isinstance(type_, sa.Time):
            raise NotImplementedError("Unsupported type: Time")

        if hints is not None:
            if dtype := hints.get(col.name):
                return dtype

        raise NotImplementedError(f"Unsupported type: {type_}")

    def replace_tbl(self, new_tbl, columns: dict[str:Column]):
        if isinstance(new_tbl, sql.Select):
            # noinspection PyNoneFunctionAssignment
            new_tbl = new_tbl.subquery()

        self.tbl = new_tbl
        self.alignment_hash = generate_alignment_hash()

        self.sql_columns = {
            col.uuid: self.tbl.columns[col.name] for col in columns.values()
        }  # from uuid to sqlalchemy column

        if hasattr(self, "cols"):
            # TODO: Clean up... This feels a bit hacky
            for col in columns.values():
                self.cols[col.uuid] = ColumnMetaData.from_expr(col.uuid, col, self)
        if hasattr(self, "intrinsic_grouped_by"):
            self.intrinsic_grouped_by.clear()

        self.joins: list[JoinDescriptor] = []
        self.wheres: list[SymbolicExpression] = []
        self.having: list[SymbolicExpression] = []
        self.order_bys: list[OrderingDescriptor] = []
        self.limit_offset: tuple[int, int] = None

    def build_select(self) -> sql.Select:
        # Validate current state
        if len(self.selects) == 0:
            raise ValueError("Can't execute a SQL query without any SELECT statements.")

        # Start building query
        select = self.tbl.select()

        # `select_from` is required if no table is explicitly referenced
        # inside the SELECT. e.g. `SELECT COUNT(*) AS count`
        select = select.select_from(self.tbl)

        # FROM
        select = self._build_select_from(select)

        # WHERE
        select = self._build_select_where(select)

        # GROUP BY
        select = self._build_select_group_by(select)

        # HAVING
        select = self._build_select_having(select)

        # LIMIT / OFFSET
        select = self._build_select_limit_offset(select)

        # SELECT
        select = self._build_select_select(select)

        # ORDER BY
        select = self._build_select_order_by(select)

        return select

    def _build_select_from(self, select):
        for join in self.joins:
            compiled, _ = self.compiler.translate(join.on, verb="join")
            on = compiled(self.sql_columns)

            select = select.join(
                join.right.tbl,
                onclause=on,
                isouter=join.how != "inner",
                full=join.how == "outer",
            )

        return select

    def _build_select_where(self, select):
        if not self.wheres:
            return select

        # Combine wheres using ands
        combined_where = functools.reduce(
            py_operator.and_, map(SymbolicExpression, self.wheres)
        )._
        compiled, where_dtype = self.compiler.translate(combined_where, verb="filter")
        assert isinstance(where_dtype, dtypes.Bool)
        where = compiled(self.sql_columns)
        return select.where(where)

    def _build_select_group_by(self, select):
        if not self.intrinsic_grouped_by:
            return select

        compiled_gb, group_by_dtypes = zip(
            *(
                self.compiler.translate(group_by, verb="group_by")
                for group_by in self.intrinsic_grouped_by
            )
        )
        group_bys = (compiled(self.sql_columns) for compiled in compiled_gb)
        return select.group_by(*group_bys)

    def _build_select_having(self, select):
        if not self.having:
            return select

        # Combine havings using ands
        combined_having = functools.reduce(
            py_operator.and_, map(SymbolicExpression, self.having)
        )._
        compiled, having_dtype = self.compiler.translate(combined_having, verb="filter")
        assert isinstance(having_dtype, dtypes.Bool)
        having = compiled(self.sql_columns)
        return select.having(having)

    def _build_select_limit_offset(self, select):
        if self.limit_offset is None:
            return select

        limit, offset = self.limit_offset
        return select.limit(limit).offset(offset)

    def _build_select_select(self, select):
        # Convert self.selects to SQLAlchemy Expressions
        s = []
        for name, uuid_ in self.selected_cols():
            sql_col = self.cols[uuid_].compiled(self.sql_columns)
            if not isinstance(sql_col, sa.sql.ColumnElement):
                sql_col = sa.literal(sql_col)
            s.append(sql_col.label(name))
        return select.with_only_columns(*s)

    def _build_select_order_by(self, select):
        if not self.order_bys:
            return select

        o = []
        for o_by in self.order_bys:
            compiled, _ = self.compiler.translate(o_by.order, verb="arrange")
            col = compiled(self.sql_columns)
            o.extend(self._order_col(col, o_by))

        return select.order_by(*o)

    #### Verb Operations ####

    def preverb_hook(self, verb: str, *args, **kwargs) -> None:
        def has_any_ftype_cols(ftypes: OPType | tuple[OPType, ...], cols: Iterable):
            if isinstance(ftypes, OPType):
                ftypes = (ftypes,)
            return any(
                self.cols[c.uuid].ftype in ftypes
                for v in cols
                for c in iterate_over_expr(self.resolve_lambda_cols(v))
                if isinstance(c, Column)
            )

        requires_subquery = False
        clear_order = False

        if self.limit_offset is not None:
            # The LIMIT / TOP clause is executed at the very end of the query.
            # This means we must create a subquery for any verb that modifies
            # the rows.
            if verb in (
                "join",
                "filter",
                "arrange",
                "group_by",
                "summarise",
            ):
                requires_subquery = True

        if verb == "mutate":
            # Window functions can't be nested, thus a subquery is required
            requires_subquery |= has_any_ftype_cols(OPType.WINDOW, kwargs.values())
        elif verb == "filter":
            # Window functions aren't allowed in where clause
            requires_subquery |= has_any_ftype_cols(OPType.WINDOW, args)
        elif verb == "summarise":
            # The result of the aggregate is always ordered according to the
            # grouping columns. We must clear the order_bys so that the order
            # is consistent with eager execution. We can do this because aggregate
            # functions are independent of the order.
            clear_order = True

            # If the grouping level is different from the grouping level of the
            # tbl object, or if on of the input columns is a window or aggregate
            # function, we must make a subquery.
            requires_subquery |= (
                bool(self.intrinsic_grouped_by)
                and self.grouped_by != self.intrinsic_grouped_by
            )
            requires_subquery |= has_any_ftype_cols(
                (OPType.AGGREGATE, OPType.WINDOW), kwargs.values()
            )

        # TODO: It would be nice if this could be done without having to select all
        #       columns. As a potential challenge for the hackathon I propose a mean
        #       of even creating the subqueries lazyly. This means that we could
        #       perform some kind of query optimization before submitting the actual
        #       query. Eg: Instead of selecting all possible columns, only select
        #       those that actually get used.
        if requires_subquery:
            columns = {
                name: self.cols[uuid].as_column(name, self)
                for name, uuid in self.named_cols.fwd.items()
            }

            original_selects = self.selects.copy()
            self.selects |= columns.keys()
            subquery = self.build_select()

            self.replace_tbl(subquery, columns)
            self.selects = original_selects

        if clear_order:
            self.order_bys.clear()

    def alias(self, name=None):
        if name is None:
            suffix = format(uuid.uuid1().int % 0x7FFFFFFF, "X")
            name = f"{self.name}_{suffix}"

        # TODO: If the table has not been modified, a simple `.alias()`
        #       would produce nicer queries.
        subquery = self.build_select().subquery(name=name)
        # In some situations sqlalchemy fails to determine the datatype of a column.
        # To circumvent this, we can pass on the information we know.
        dtype_hints = {
            name: self.cols[self.named_cols.fwd[name]].dtype for name in self.selects
        }

        return self.__class__(self.engine, subquery, _dtype_hints=dtype_hints)

    def collect(self):
        select = self.build_select()
        with self.engine.connect() as conn:
            try:
                # TODO: check for which pandas versions this is needed:
                # Temporary fix for pandas bug (https://github.com/pandas-dev/pandas/issues/35484)
                # Taken from siuba
                from pandas.io import sql as _pd_sql

                class _FixedSqlDatabase(_pd_sql.SQLDatabase):
                    def execute(self, *args, **kwargs):
                        return self.connectable.execute(*args, **kwargs)

                sql_db = _FixedSqlDatabase(conn)
                result = sql_db.read_sql(select).convert_dtypes()
            except AttributeError:
                import pandas as pd

                result = pd.read_sql_query(select, con=conn)

        # Add metadata
        result.attrs["name"] = self.name
        return result

    def build_query(self) -> str:
        query = self.build_select()
        return str(
            query.compile(
                dialect=self.engine.dialect, compile_kwargs={"literal_binds": True}
            )
        )

    def join(self, right, on, how, *, validate=None):
        self.alignment_hash = generate_alignment_hash()

        # If right has joins already, merging them becomes extremely difficult
        # This is because the ON clauses could contain NULL checks in which case
        # the joins aren't associative anymore.
        if right.joins:
            raise ValueError(
                "Can't automatically combine joins if the right side already contains a"
                " JOIN clause."
            )

        if right.limit_offset is not None:
            raise ValueError(
                "The right table can't be sliced when performing a join."
                " Wrap the right side in a subquery to fix this."
            )

        # TODO: Handle GROUP BY and SELECTS on left / right side

        # Combine the WHERE clauses
        if how == "inner":
            # Inner Join: The WHERES can be combined
            self.wheres.extend(right.wheres)
        elif how == "left":
            # WHERES from right must go into the ON clause
            on = reduce(py_operator.and_, (on, *right.wheres))
        elif how == "outer":
            # For outer joins, the WHERE clause can't easily be merged.
            # The best solution for now is to move them into a subquery.
            if self.wheres:
                raise ValueError(
                    "Filters can't precede outer joins. Wrap the left side in a"
                    " subquery to fix this."
                )
            if right.wheres:
                raise ValueError(
                    "Filters can't precede outer joins. Wrap the right side in a"
                    " subquery to fix this."
                )

        if validate is not None:
            warnings.warn("SQL table backend ignores join validation argument.")

        descriptor = JoinDescriptor(right, on, how)
        self.joins.append(descriptor)

        self.sql_columns.update(right.sql_columns)

    def filter(self, *args):
        self.alignment_hash = generate_alignment_hash()

        if self.intrinsic_grouped_by:
            for arg in args:
                # If a condition involves only grouping columns, it can be
                # moved into the wheres instead of the havings.
                only_grouping_cols = all(
                    col in self.intrinsic_grouped_by
                    for col in iterate_over_expr(arg, expand_literal_col=True)
                    if isinstance(col, Column)
                )

                if only_grouping_cols:
                    self.wheres.append(arg)
                else:
                    self.having.append(arg)
        else:
            self.wheres.extend(args)

    def arrange(self, ordering):
        self.alignment_hash = generate_alignment_hash()

        # Merge order bys and remove duplicate columns
        order_bys = []
        order_by_columns = set()
        for o_by in ordering + self.order_bys:
            if o_by.order in order_by_columns:
                continue
            order_bys.append(o_by)
            order_by_columns.add(o_by.order)

        self.order_bys = order_bys

    def summarise(self, **kwargs):
        self.alignment_hash = generate_alignment_hash()

    def slice_head(self, n: int, offset: int):
        if self.limit_offset is None:
            self.limit_offset = (n, offset)
        else:
            old_n, old_o = self.limit_offset
            self.limit_offset = (min(abs(old_n - offset), n), old_o + offset)

    #### EXPRESSIONS ####

    def _order_col(
        self, col: sa.SQLColumnExpression, ordering: OrderingDescriptor
    ) -> list[sa.SQLColumnExpression]:
        col = col.asc() if ordering.asc else col.desc()
        col = col.nullsfirst() if ordering.nulls_first else col.nullslast()
        return [col]

    class ExpressionCompiler(
        LazyTableImpl.ExpressionCompiler[
            "SQLTableImpl",
            TypedValue[Callable[[dict[uuid.UUID, sa.Column]], sql.ColumnElement]],
        ]
    ):
        def _translate_col(self, expr, **kwargs):
            # Can either be a base SQL column, or a reference to an expression
            if expr.uuid in self.backend.sql_columns:

                def sql_col(cols, **kw):
                    return cols[expr.uuid]

                return TypedValue(sql_col, expr.dtype, OPType.EWISE)

            col = self.backend.cols[expr.uuid]
            return TypedValue(col.compiled, col.dtype, col.ftype)

        def _translate_literal_col(self, expr, **kwargs):
            if not self.backend.is_aligned_with(expr):
                raise AlignmentError(
                    "Literal column isn't aligned with this table. "
                    f"Literal Column: {expr}"
                )

            def sql_col(cols, **kw):
                return expr.typed_value.value

            return TypedValue(sql_col, expr.typed_value.dtype, expr.typed_value.ftype)

        def _translate_function(
            self, expr, implementation, op_args, context_kwargs, *, verb=None, **kwargs
        ):
            value = self._translate_function_value(
                expr,
                implementation,
                op_args,
                context_kwargs,
                verb=verb,
                **kwargs,
            )
            operator = implementation.operator

            if operator.ftype == OPType.AGGREGATE and verb == "mutate":
                # Aggregate function in mutate verb -> window function
                over_value = self.over_clause(value, implementation, context_kwargs)
                ftype = self.backend._get_op_ftype(
                    op_args, operator, OPType.WINDOW, strict=True
                )
                return TypedValue(over_value, implementation.rtype, ftype)

            elif operator.ftype == OPType.WINDOW:
                if verb != "mutate":
                    raise FunctionTypeError(
                        "Window function are only allowed inside a mutate."
                    )

                over_value = self.over_clause(value, implementation, context_kwargs)
                ftype = self.backend._get_op_ftype(op_args, operator, strict=True)
                return TypedValue(over_value, implementation.rtype, ftype)

            else:
                ftype = self.backend._get_op_ftype(op_args, operator, strict=True)
                return TypedValue(value, implementation.rtype, ftype)

        def _translate_function_value(
            self,
            expr: FunctionCall,
            implementation: TypedOperatorImpl,
            op_args: list,
            context_kwargs: dict,
            *,
            verb=None,
            **kwargs,
        ):
            impl_dtypes = implementation.impl.signature.args
            if implementation.impl.signature.is_vararg:
                impl_dtypes = itertools.chain(
                    impl_dtypes[:-1],
                    itertools.repeat(impl_dtypes[-1]),
                )

            def value(cols, *, variant=None, internal_kwargs=None, **kw):
                args = []
                for arg, dtype in zip(op_args, impl_dtypes):
                    if dtype.const:
                        args.append(arg.value(cols, as_sql_literal=False))
                    else:
                        args.append(arg.value(cols))

                kwargs = {
                    "_tbl": self.backend,
                    "_verb": verb,
                    **(internal_kwargs or {}),
                }

                if variant is not None:
                    if variant_impl := implementation.get_variant(variant):
                        return variant_impl(*args, **kwargs)

                return implementation(*args, **kwargs)

            return value

        def _translate_case(self, expr, switching_on, cases, default, **kwargs):
            def value(*args, **kw):
                default_ = default.value(*args, **kwargs)

                if switching_on is not None:
                    switching_on_ = switching_on.value(*args, **kwargs)
                    return sa.case(
                        {
                            cond.value(*args, **kw): val.value(*args, **kw)
                            for cond, val in cases
                        },
                        value=switching_on_,
                        else_=default_,
                    )

                return sa.case(
                    *(
                        (cond.value(*args, **kw), val.value(*args, **kw))
                        for cond, val in cases
                    ),
                    else_=default_,
                )

            result_dtype, result_ftype = self._translate_case_common(
                expr, switching_on, cases, default, **kwargs
            )
            return TypedValue(value, result_dtype, result_ftype)

        def _translate_literal_value(self, expr):
            def literal_func(*args, as_sql_literal=True, **kwargs):
                if as_sql_literal:
                    return sa.literal(expr)
                return expr

            return literal_func

        def over_clause(
            self,
            value: Callable,
            implementation: TypedOperatorImpl,
            context_kwargs: dict,
        ):
            operator = implementation.operator
            if operator.ftype not in (OPType.AGGREGATE, OPType.WINDOW):
                raise FunctionTypeError

            wants_order_by = operator.ftype == OPType.WINDOW

            # PARTITION BY
            compiled_pb = tuple(
                self.translate(group_by).value for group_by in self.backend.grouped_by
            )

            # ORDER BY
            def order_by_clause_generator(ordering: OrderingDescriptor):
                compiled, _ = self.translate(ordering.order)

                def clause(*args, **kwargs):
                    col = compiled(*args, **kwargs)
                    return self.backend._order_col(col, ordering)

                return clause

            if wants_order_by:
                arrange = context_kwargs.get("arrange")
                if not arrange:
                    raise TypeError("Missing 'arrange' argument.")

                ordering = translate_ordering(self.backend, arrange)
                compiled_ob = [order_by_clause_generator(o_by) for o_by in ordering]

            # New value callable
            def over_value(*args, **kwargs):
                pb = sql.expression.ClauseList(
                    *(compiled(*args, **kwargs) for compiled in compiled_pb)
                )
                ob = (
                    sql.expression.ClauseList(
                        *(
                            clause
                            for compiled in compiled_ob
                            for clause in compiled(*args, **kwargs)
                        )
                    )
                    if wants_order_by
                    else None
                )

                # Some operators need to further modify the OVER expression
                # To do this, we allow registering a variant called "window"
                if implementation.has_variant("window"):
                    return value(
                        *args,
                        variant="window",
                        internal_kwargs={
                            "_window_partition_by": pb,
                            "_window_order_by": ob,
                        },
                        **kwargs,
                    )

                # If now window variant has been defined, just apply generic OVER clause
                return value(*args, **kwargs).over(
                    partition_by=pb,
                    order_by=ob,
                )

            return over_value

    class AlignedExpressionEvaluator(
        LazyTableImpl.AlignedExpressionEvaluator[TypedValue[sql.ColumnElement]]
    ):
        def translate(self, expr, check_alignment=True, **kwargs):
            if check_alignment:
                alignment_hashes = {
                    col.table.alignment_hash
                    for col in iterate_over_expr(expr, expand_literal_col=True)
                    if isinstance(col, Column)
                }
                if len(alignment_hashes) >= 2:
                    raise AlignmentError(
                        "Expression contains columns from different tables that aren't"
                        " aligned."
                    )

            return super().translate(expr, check_alignment=check_alignment, **kwargs)

        def _translate_col(self, expr, **kwargs):
            backend = expr.table
            if expr.uuid in backend.sql_columns:
                sql_col = backend.sql_columns[expr.uuid]
                return TypedValue(sql_col, expr.dtype)

            col = backend.cols[expr.uuid]
            return TypedValue(col.compiled(backend.sql_columns), col.dtype, col.ftype)

        def _translate_literal_col(self, expr, **kwargs):
            assert issubclass(expr.backend, SQLTableImpl)
            return expr.typed_value

        def _translate_function(
            self, expr, implementation, op_args, context_kwargs, **kwargs
        ):
            # Aggregate function -> window function
            value = implementation(*(arg.value for arg in op_args))
            operator = implementation.operator
            override_ftype = (
                OPType.WINDOW if operator.ftype == OPType.AGGREGATE else None
            )
            ftype = SQLTableImpl._get_op_ftype(
                op_args, operator, override_ftype, strict=True
            )

            if operator.ftype == OPType.AGGREGATE:
                value = value.over()
            if operator.ftype == OPType.WINDOW:
                raise NotImplementedError("How to handle window functions?")

            return TypedValue(value, implementation.rtype, ftype)


def generate_alignment_hash():
    # It should be possible to have an alternative hash value that
    # is a bit more lenient -> If the same set of operations get applied
    # to a table in two different orders that produce the same table
    # object, their hash could also be equal.
    return uuid.uuid1()


#### BACKEND SPECIFIC OPERATORS ################################################


with SQLTableImpl.op(ops.FloorDiv(), check_super=False) as op:
    if sa.__version__ < "2":

        @op.auto
        def _floordiv(lhs, rhs):
            return sa.cast(lhs / rhs, sa.Integer())

    else:

        @op.auto
        def _floordiv(lhs, rhs):
            return lhs // rhs


with SQLTableImpl.op(ops.RFloorDiv(), check_super=False) as op:

    @op.auto
    def _rfloordiv(rhs, lhs):
        return _floordiv(lhs, rhs)


with SQLTableImpl.op(ops.Pow()) as op:

    @op.auto
    def _pow(lhs, rhs):
        if isinstance(lhs.type, sa.Float) or isinstance(rhs.type, sa.Float):
            type_ = sa.Double()
        elif isinstance(lhs.type, sa.Numeric) or isinstance(rhs, sa.Numeric):
            type_ = sa.Numeric()
        else:
            type_ = sa.Double()

        return sa.func.POW(lhs, rhs, type_=type_)


with SQLTableImpl.op(ops.RPow()) as op:

    @op.auto
    def _rpow(rhs, lhs):
        return _pow(lhs, rhs)


with SQLTableImpl.op(ops.Xor()) as op:

    @op.auto
    def _xor(lhs, rhs):
        return lhs != rhs


with SQLTableImpl.op(ops.RXor()) as op:

    @op.auto
    def _rxor(rhs, lhs):
        return lhs != rhs


with SQLTableImpl.op(ops.Pos()) as op:

    @op.auto
    def _pos(x):
        return x


with SQLTableImpl.op(ops.Abs()) as op:

    @op.auto
    def _abs(x):
        return sa.func.ABS(x, type_=x.type)


with SQLTableImpl.op(ops.Round()) as op:

    @op.auto
    def _round(x, decimals=0):
        return sa.func.ROUND(x, decimals, type_=x.type)


with SQLTableImpl.op(ops.IsIn()) as op:

    @op.auto
    def _isin(x, *values, _verb=None):
        if _verb == "filter":
            # In WHERE and HAVING clause, we can use the IN operator
            return x.in_(values)
        # In SELECT we must replace it with the corresponding boolean expression
        return reduce(py_operator.or_, map(lambda v: x == v, values))


#### String Functions ####


with SQLTableImpl.op(ops.Strip()) as op:

    @op.auto
    def _strip(x):
        return sa.func.TRIM(x, type_=x.type)


with SQLTableImpl.op(ops.StringLength()) as op:

    @op.auto
    def _str_length(x):
        return sa.func.LENGTH(x, type_=sa.Integer())


with SQLTableImpl.op(ops.Upper()) as op:

    @op.auto
    def _upper(x):
        return sa.func.UPPER(x, type_=x.type)


with SQLTableImpl.op(ops.Lower()) as op:

    @op.auto
    def _upper(x):
        return sa.func.LOWER(x, type_=x.type)


with SQLTableImpl.op(ops.Replace()) as op:

    @op.auto
    def _replace(x, y, z):
        return sa.func.REPLACE(x, y, z, type_=x.type)


with SQLTableImpl.op(ops.StartsWith()) as op:

    @op.auto
    def _startswith(x, y):
        return x.startswith(y, autoescape=True)


with SQLTableImpl.op(ops.EndsWith()) as op:

    @op.auto
    def _endswith(x, y):
        return x.endswith(y, autoescape=True)


with SQLTableImpl.op(ops.Contains()) as op:

    @op.auto
    def _contains(x, y):
        return x.contains(y, autoescape=True)


#### Datetime Functions ####


with SQLTableImpl.op(ops.Year()) as op:

    @op.auto
    def _year(x):
        return sa.extract("year", x)


with SQLTableImpl.op(ops.Month()) as op:

    @op.auto
    def _month(x):
        return sa.extract("month", x)


with SQLTableImpl.op(ops.Day()) as op:

    @op.auto
    def _day(x):
        return sa.extract("day", x)


with SQLTableImpl.op(ops.Hour()) as op:

    @op.auto
    def _hour(x):
        return sa.extract("hour", x)


with SQLTableImpl.op(ops.Minute()) as op:

    @op.auto
    def _minute(x):
        return sa.extract("minute", x)


with SQLTableImpl.op(ops.Second()) as op:

    @op.auto
    def _second(x):
        return sa.extract("second", x)


with SQLTableImpl.op(ops.Millisecond()) as op:

    @op.auto
    def _millisecond(x):
        return sa.extract("milliseconds", x) % 1000


with SQLTableImpl.op(ops.DayOfWeek()) as op:

    @op.auto
    def _day_of_week(x):
        return sa.extract("dow", x)


with SQLTableImpl.op(ops.DayOfYear()) as op:

    @op.auto
    def _day_of_year(x):
        return sa.extract("doy", x)


#### Generic Functions ####


with SQLTableImpl.op(ops.Greatest()) as op:

    @op.auto
    def _greatest(*x):
        # TODO: Determine return type
        return sa.func.GREATEST(*x)


with SQLTableImpl.op(ops.Least()) as op:

    @op.auto
    def _least(*x):
        # TODO: Determine return type
        return sa.func.LEAST(*x)


#### Summarising Functions ####


with SQLTableImpl.op(ops.Mean()) as op:

    @op.auto
    def _mean(x):
        type_ = sa.Numeric()
        if isinstance(x.type, sa.Float):
            type_ = sa.Double()

        return sa.func.AVG(x, type_=type_)


with SQLTableImpl.op(ops.Min()) as op:

    @op.auto
    def _min(x):
        return sa.func.min(x)


with SQLTableImpl.op(ops.Max()) as op:

    @op.auto
    def _max(x):
        return sa.func.max(x)


with SQLTableImpl.op(ops.Sum()) as op:

    @op.auto
    def _sum(x):
        return sa.func.sum(x)


with SQLTableImpl.op(ops.Any()) as op:

    @op.auto
    def _any(x, *, _window_partition_by=None, _window_order_by=None):
        return sa.func.coalesce(sa.func.max(x), sa.false())

    @op.auto(variant="window")
    def _any(x, *, _window_partition_by=None, _window_order_by=None):
        return sa.func.coalesce(
            sa.func.max(x).over(
                partition_by=_window_partition_by,
                order_by=_window_order_by,
            ),
            sa.false(),
        )


with SQLTableImpl.op(ops.All()) as op:

    @op.auto
    def _all(x):
        return sa.func.coalesce(sa.func.min(x), sa.false())

    @op.auto(variant="window")
    def _all(x, *, _window_partition_by=None, _window_order_by=None):
        return sa.func.coalesce(
            sa.func.min(x).over(
                partition_by=_window_partition_by,
                order_by=_window_order_by,
            ),
            sa.false(),
        )


with SQLTableImpl.op(ops.StringJoin()) as op:

    @op.auto
    def _join(x, sep: str):
        return sa.func.STRING_AGG(x, sep, type_=x.type)


with SQLTableImpl.op(ops.Count()) as op:

    @op.auto
    def _count(x=None):
        if x is None:
            # Get the number of rows
            return sa.func.count()
        else:
            # Count non null values
            return sa.func.count(x)


#### Window Functions ####


with SQLTableImpl.op(ops.Shift()) as op:

    @op.auto
    def _shift():
        raise RuntimeError("This is a stub")

    @op.auto(variant="window")
    def _shift(
        x,
        by,
        empty_value=None,
        *,
        _window_partition_by=None,
        _window_order_by=None,
    ):
        if by == 0:
            return x
        if by > 0:
            return sa.func.LAG(x, by, empty_value, type_=x.type).over(
                partition_by=_window_partition_by, order_by=_window_order_by
            )
        if by < 0:
            return sa.func.LEAD(x, -by, empty_value, type_=x.type).over(
                partition_by=_window_partition_by, order_by=_window_order_by
            )


with SQLTableImpl.op(ops.RowNumber()) as op:

    @op.auto
    def _row_number():
        return sa.func.ROW_NUMBER(type_=sa.Integer())


with SQLTableImpl.op(ops.Rank()) as op:

    @op.auto
    def _rank(_):
        return sa.func.rank()


with SQLTableImpl.op(ops.DenseRank()) as op:

    @op.auto
    def _dense_rank(_):
        return sa.func.dense_rank()


####


# Load SQLTableImpl dialect specific subclasses
from . import dialects  # noqa



from __future__ import annotations

import copy
import dataclasses
import datetime
import uuid
import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

from pydiverse.transform import ops
from pydiverse.transform._typing import ImplT
from pydiverse.transform.core import dtypes
from pydiverse.transform.core.expressions import (
    CaseExpression,
    Column,
    LambdaColumn,
    LiteralColumn,
)
from pydiverse.transform.core.expressions.translator import (
    DelegatingTranslator,
    Translator,
    TypedValue,
)
from pydiverse.transform.core.registry import (
    OperatorRegistrationContextManager,
    OperatorRegistry,
)
from pydiverse.transform.core.util import bidict, ordered_set
from pydiverse.transform.errors import ExpressionTypeError, FunctionTypeError
from pydiverse.transform.ops import OPType

if TYPE_CHECKING:
    from pydiverse.transform.core.util import OrderingDescriptor
    from pydiverse.transform.ops import Operator


ExprCompT = TypeVar("ExprCompT", bound="TypedValue")
AlignedT = TypeVar("AlignedT", bound="TypedValue")


class AbstractTableImpl:
    """
    Base class from which all table backend implementations are derived from.
    It tracks various metadata that is relevant for all backends.

    Attributes:
        name: The name of the table.

        selects: Ordered set of selected names.
        named_cols: Map from name to column uuid containing all columns that
            have been named.
        available_cols: Set of UUIDs that can be referenced in symbolic
            expressions. This set gets used to validate verb inputs. It usually
            contains the same uuids as the col_exprs. Only a summarising
            operation resets this.
        col_expr: Map from uuid to the `SymbolicExpression` that corresponds
            to this column.
        col_dtype: Map from uuid to the datatype of the corresponding column.
            It is the responsibility of the backend to keep track of
            this information.

        grouped_by: Ordered set of columns by which the table is grouped by.
        intrinsic_grouped_by: Ordered set of columns representing the underlying
            grouping level of the table. This gets set when performing a
            summarising operation.
    """

    operator_registry = OperatorRegistry("AbstractTableImpl")

    def __init__(
        self,
        name: str,
        columns: dict[str, Column],
    ):
        self.name = name
        self.compiler = self.ExpressionCompiler(self)
        self.lambda_translator = self.LambdaTranslator(self)

        self.selects = ordered_set()  # type: ordered_set[str]
        self.named_cols = bidict()  # type: bidict[str: uuid.UUID]
        self.available_cols = set()  # type: set[uuid.UUID]
        self.cols = {}  # type: dict[uuid.UUID: ColumnMetaData]

        self.grouped_by = ordered_set()  # type: ordered_set[Column]
        self.intrinsic_grouped_by = ordered_set()  # type: ordered_set[Column]

        # Init Values
        for name, col in columns.items():
            self.selects.add(name)
            self.named_cols.fwd[name] = col.uuid
            self.available_cols.add(col.uuid)
            self.cols[col.uuid] = ColumnMetaData.from_expr(col.uuid, col, self)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Add new `operator_registry` class variable to subclass.
        # We define the super registry by walking up the MRO. This allows us
        # to check for potential operation definitions in the parent classes.
        super_reg = None
        for super_cls in cls.__mro__:
            if hasattr(super_cls, "operator_registry"):
                super_reg = super_cls.operator_registry
                break
        cls.operator_registry = OperatorRegistry(cls.__name__, super_reg)

    def copy(self):
        c = copy.copy(self)
        # Copy containers
        for k, v in self.__dict__.items():
            if isinstance(v, (list, dict, set, bidict, ordered_set)):
                c.__dict__[k] = copy.copy(v)

        # Must create a new translator, so that it can access the current df.
        c.compiler = self.ExpressionCompiler(c)
        c.lambda_translator = self.LambdaTranslator(c)
        return c

    def get_col(self, key: str | Column | LambdaColumn):
        """Getter used by `Table.__getattr__`"""

        if isinstance(key, LambdaColumn):
            key = key.name

        if isinstance(key, str):
            if uuid := self.named_cols.fwd.get(key, None):
                return self.cols[uuid].as_column(key, self)
            # Must return AttributeError, else `hasattr` doesn't work on Table instances
            raise AttributeError(f"Table '{self.name}' has not column named '{key}'.")

        if isinstance(key, Column):
            uuid = key.uuid
            if uuid in self.available_cols:
                name = self.named_cols.bwd[uuid]
                return self.cols[uuid].as_column(name, self)
            raise KeyError(f"Table '{self.name}' has no column that matches '{key}'.")

    def selected_cols(self) -> Iterable[tuple[str, uuid.UUID]]:
        for name in self.selects:
            yield (name, self.named_cols.fwd[name])

    def resolve_lambda_cols(self, expr: Any):
        return self.lambda_translator.translate(expr)

    def is_aligned_with(self, col: Column | LiteralColumn) -> bool:
        """Determine if a column is aligned with the table.

        :param col: The column or literal colum against which alignment
            should be checked.
        :return: A boolean indicating if `col` is aligned with self.
        """
        raise NotImplementedError

    @classmethod
    def _html_repr_expr(cls, expr):
        """
        Return an appropriate string to display an expression from this backend.
        This is mainly used to IPython.
        """
        return repr(expr)

    #### Verb Callbacks ####

    def preverb_hook(self, verb: str, *args, **kwargs) -> None:
        """Hook that gets called right after `copy` inside a verb

        This gives the backend a chance to react and modify it's state. This
        can, for example, be used to create a subquery based on specific
        conditions.

        :param verb: The name of the verb
        :param args: The arguments passed to the verb
        :param kwargs: The keyword arguments passed to the verb
        """
        ...

    def alias(self, name=None) -> AbstractTableImpl:
        ...

    def collect(self):
        ...

    def build_query(self):
        ...

    def select(self, *args):
        ...

    def mutate(self, **kwargs):
        ...

    def join(self, right, on, how, *, validate=None):
        ...

    def filter(self, *args):
        ...

    def arrange(self, ordering: list[OrderingDescriptor]):
        ...

    def group_by(self, *args):
        ...

    def ungroup(self, *args):
        ...

    def summarise(self, **kwargs):
        ...

    def slice_head(self, n: int, offset: int):
        ...

    #### Symbolic Operators ####

    @classmethod
    def op(cls, operator: Operator, **kwargs) -> OperatorRegistrationContextManager:
        return OperatorRegistrationContextManager(
            cls.operator_registry, operator, **kwargs
        )

    #### Expressions ####

    class ExpressionCompiler(
        DelegatingTranslator[ExprCompT], Generic[ImplT, ExprCompT]
    ):
        """
        Class convert an expression into a function that, when provided with
        the appropriate arguments, evaluates the expression.

        The reason we can't just eagerly evaluate the expression is because for
        grouped data we often have to use the split-apply-combine strategy.
        """

        def __init__(self, backend: ImplT):
            self.backend = backend
            super().__init__(backend.operator_registry)

        def _translate_literal(self, expr, **kwargs):
            literal = self._translate_literal_value(expr)

            if isinstance(expr, bool):
                return TypedValue(literal, dtypes.Bool(const=True))
            if isinstance(expr, int):
                return TypedValue(literal, dtypes.Int(const=True))
            if isinstance(expr, float):
                return TypedValue(literal, dtypes.Float(const=True))
            if isinstance(expr, str):
                return TypedValue(literal, dtypes.String(const=True))
            if isinstance(expr, datetime.datetime):
                return TypedValue(literal, dtypes.DateTime(const=True))

            if expr is None:
                return TypedValue(literal, dtypes.NoneDType(const=True))

        def _translate_literal_value(self, expr):
            def literal_func(*args, **kwargs):
                return expr

            return literal_func

        def _translate_case_common(
            self,
            expr: CaseExpression,
            switching_on: ExprCompT | None,
            cases: list[tuple[ExprCompT, ExprCompT]],
            default: ExprCompT,
            **kwargs,
        ) -> tuple[dtypes.DType, OPType]:
            # Determine dtype of result
            val_dtypes = [default.dtype.without_modifiers()]
            for _, val in cases:
                val_dtypes.append(val.dtype.without_modifiers())

            result_dtype = dtypes.promote_dtypes(val_dtypes)

            # Determine ftype of result
            val_ftypes = set()
            if not default.dtype.const:
                val_ftypes.add(default.ftype)

            for _, val in cases:
                if not val.dtype.const:
                    val_ftypes.add(val.ftype)

            if len(val_ftypes) == 0:
                result_ftype = OPType.EWISE
            elif len(val_ftypes) == 1:
                (result_ftype,) = val_ftypes
            elif OPType.WINDOW in val_ftypes:
                result_ftype = OPType.WINDOW
            else:
                # AGGREGATE and EWISE are incompatible
                raise FunctionTypeError(
                    "Incompatible function types found in case statement: "
                    ", ".join(val_ftypes)
                )

            if result_ftype is OPType.EWISE and switching_on is not None:
                result_ftype = switching_on.ftype

            # Type check conditions
            if switching_on is None:
                # All conditions must be boolean
                for cond, _ in cases:
                    if not dtypes.Bool().same_kind(cond.dtype):
                        raise ExpressionTypeError(
                            "All conditions in a case statement return booleans. "
                            f"{cond} is of type {cond.dtype}."
                        )
            else:
                # All conditions must be of the same type as switching_on
                for cond, _ in cases:
                    if not cond.dtype.can_promote_to(
                        switching_on.dtype.without_modifiers()
                    ):
                        # Can't compare
                        raise ExpressionTypeError(
                            f"Condition value {cond} (dtype: {cond.dtype}) "
                            f"is incompatible with switch dtype {switching_on.dtype}."
                        )

            return result_dtype, result_ftype

    class AlignedExpressionEvaluator(DelegatingTranslator[AlignedT], Generic[AlignedT]):
        """
        Used for evaluating an expression in a typical eager style where, as
        long as two columns have the same alignment / length, we can perform
        operations on them without first having to join them.
        """

        def _translate_literal(self, expr, **kwargs):
            if isinstance(expr, bool):
                return TypedValue(expr, dtypes.Bool(const=True))
            if isinstance(expr, int):
                return TypedValue(expr, dtypes.Int(const=True))
            if isinstance(expr, float):
                return TypedValue(expr, dtypes.Float(const=True))
            if isinstance(expr, str):
                return TypedValue(expr, dtypes.String(const=True))
            if isinstance(expr, datetime.datetime):
                return TypedValue(expr, dtypes.DateTime(const=True))

            if expr is None:
                return TypedValue(expr, dtypes.NoneDType(const=True))

    class LambdaTranslator(Translator):
        """
        Translator that takes an expression and replaces all LambdaColumns
        inside it with the corresponding Column instance.
        """

        def __init__(self, backend: ImplT):
            self.backend = backend
            super().__init__()

        def _translate(self, expr, **kwargs):
            # Resolve lambda and return Column object
            if isinstance(expr, LambdaColumn):
                if expr.name not in self.backend.named_cols.fwd:
                    raise ValueError(
                        f"Invalid lambda column '{expr.name}'. No column with this name"
                        f" found for table '{self.backend.name}'."
                    )
                uuid = self.backend.named_cols.fwd[expr.name]
                return self.backend.cols[uuid].as_column(expr.name, self.backend)
            return expr

    #### Helpers ####

    @classmethod
    def _get_op_ftype(
        cls, args, operator: Operator, override_ftype: OPType = None, strict=False
    ) -> OPType:
        """
        Get the ftype based on a function implementation and the arguments.

            e(e) -> e       a(e) -> a       w(e) -> w
            e(a) -> a       a(a) -> Err     w(a) -> w
            e(w) -> w       a(w) -> Err     w(w) -> Err

        If the implementation ftype is incompatible with the arguments, this
        function raises an Exception.
        """

        ftypes = [arg.ftype for arg in args]
        op_ftype = override_ftype or operator.ftype

        if op_ftype == OPType.EWISE:
            if OPType.WINDOW in ftypes:
                return OPType.WINDOW
            if OPType.AGGREGATE in ftypes:
                return OPType.AGGREGATE
            return op_ftype

        if op_ftype == OPType.AGGREGATE:
            if OPType.WINDOW in ftypes:
                if strict:
                    raise FunctionTypeError(
                        "Can't nest a window function inside an aggregate function"
                        f" ({operator.name})."
                    )
                else:
                    # TODO: Replace with logger
                    warnings.warn(
                        "Nesting a window function inside an aggregate function is not"
                        " supported by SQL backend."
                    )
            if OPType.AGGREGATE in ftypes:
                raise FunctionTypeError(
                    "Can't nest an aggregate function inside an aggregate function"
                    f" ({operator.name})."
                )
            return op_ftype

        if op_ftype == OPType.WINDOW:
            if OPType.WINDOW in ftypes:
                if strict:
                    raise FunctionTypeError(
                        "Can't nest a window function inside a window function"
                        f" ({operator.name})."
                    )
                else:
                    warnings.warn(
                        "Nesting a window function inside a window function is not"
                        " supported by SQL backend."
                    )
            return op_ftype


@dataclasses.dataclass
class ColumnMetaData:
    uuid: uuid.UUID
    expr: Any
    compiled: Callable[[Any], TypedValue]
    dtype: dtypes.DType
    ftype: OPType

    @classmethod
    def from_expr(cls, uuid, expr, table: AbstractTableImpl, **kwargs):
        v = table.compiler.translate(expr, **kwargs)
        return cls(
            uuid=uuid,
            expr=expr,
            compiled=v.value,
            dtype=v.dtype.without_modifiers(),
            ftype=v.ftype,
        )

    def __hash__(self):
        return hash(self.uuid)

    def as_column(self, name, table: AbstractTableImpl):
        return Column(name, table, self.dtype, self.uuid)


#### MARKER OPERATIONS #########################################################


with AbstractTableImpl.op(ops.NullsFirst()) as op:

    @op.auto
    def _nulls_first(_):
        raise RuntimeError("This is just a marker that never should get called")


with AbstractTableImpl.op(ops.NullsLast()) as op:

    @op.auto
    def _nulls_last(_):
        raise RuntimeError("This is just a marker that never should get called")


#### ARITHMETIC OPERATORS ######################################################


with AbstractTableImpl.op(ops.Add()) as op:

    @op.auto
    def _add(lhs, rhs):
        return lhs + rhs

    @op.extension(ops.StringAdd)
    def _str_add(lhs, rhs):
        return lhs + rhs


with AbstractTableImpl.op(ops.RAdd()) as op:

    @op.auto
    def _radd(rhs, lhs):
        return lhs + rhs

    @op.extension(ops.StringRAdd)
    def _str_radd(lhs, rhs):
        return lhs + rhs


with AbstractTableImpl.op(ops.Sub()) as op:

    @op.auto
    def _sub(lhs, rhs):
        return lhs - rhs


with AbstractTableImpl.op(ops.RSub()) as op:

    @op.auto
    def _rsub(rhs, lhs):
        return lhs - rhs


with AbstractTableImpl.op(ops.Mul()) as op:

    @op.auto
    def _mul(lhs, rhs):
        return lhs * rhs


with AbstractTableImpl.op(ops.RMul()) as op:

    @op.auto
    def _rmul(rhs, lhs):
        return lhs * rhs


with AbstractTableImpl.op(ops.TrueDiv()) as op:

    @op.auto
    def _truediv(lhs, rhs):
        return lhs / rhs


with AbstractTableImpl.op(ops.RTrueDiv()) as op:

    @op.auto
    def _rtruediv(rhs, lhs):
        return lhs / rhs


with AbstractTableImpl.op(ops.FloorDiv()) as op:

    @op.auto
    def _floordiv(lhs, rhs):
        return lhs // rhs


with AbstractTableImpl.op(ops.RFloorDiv()) as op:

    @op.auto
    def _rfloordiv(rhs, lhs):
        return lhs // rhs


with AbstractTableImpl.op(ops.Pow()) as op:

    @op.auto
    def _pow(lhs, rhs):
        return lhs**rhs


with AbstractTableImpl.op(ops.RPow()) as op:

    @op.auto
    def _rpow(rhs, lhs):
        return lhs**rhs


with AbstractTableImpl.op(ops.Mod()) as op:

    @op.auto
    def _mod(lhs, rhs):
        return lhs % rhs


with AbstractTableImpl.op(ops.RMod()) as op:

    @op.auto
    def _rmod(rhs, lhs):
        return lhs % rhs


with AbstractTableImpl.op(ops.Neg()) as op:

    @op.auto
    def _neg(x):
        return -x


with AbstractTableImpl.op(ops.Pos()) as op:

    @op.auto
    def _pos(x):
        return +x


with AbstractTableImpl.op(ops.Abs()) as op:

    @op.auto
    def _abs(x):
        return abs(x)


#### BINARY OPERATORS ##########################################################


with AbstractTableImpl.op(ops.And()) as op:

    @op.auto
    def _and(lhs, rhs):
        return lhs & rhs


with AbstractTableImpl.op(ops.RAnd()) as op:

    @op.auto
    def _rand(rhs, lhs):
        return lhs & rhs


with AbstractTableImpl.op(ops.Or()) as op:

    @op.auto
    def _or(lhs, rhs):
        return lhs | rhs


with AbstractTableImpl.op(ops.ROr()) as op:

    @op.auto
    def _ror(rhs, lhs):
        return lhs | rhs


with AbstractTableImpl.op(ops.Xor()) as op:

    @op.auto
    def _xor(lhs, rhs):
        return lhs ^ rhs


with AbstractTableImpl.op(ops.RXor()) as op:

    @op.auto
    def _rxor(rhs, lhs):
        return lhs ^ rhs


with AbstractTableImpl.op(ops.Invert()) as op:

    @op.auto
    def _invert(x):
        return ~x


#### COMPARISON OPERATORS ######################################################


with AbstractTableImpl.op(ops.Equal()) as op:

    @op.auto
    def _eq(lhs, rhs):
        return lhs == rhs


with AbstractTableImpl.op(ops.NotEqual()) as op:

    @op.auto
    def _ne(lhs, rhs):
        return lhs != rhs


with AbstractTableImpl.op(ops.Less()) as op:

    @op.auto
    def _lt(lhs, rhs):
        return lhs < rhs


with AbstractTableImpl.op(ops.LessEqual()) as op:

    @op.auto
    def _le(lhs, rhs):
        return lhs <= rhs


with AbstractTableImpl.op(ops.Greater()) as op:

    @op.auto
    def _gt(lhs, rhs):
        return lhs > rhs


with AbstractTableImpl.op(ops.GreaterEqual()) as op:

    @op.auto
    def _ge(lhs, rhs):
        return lhs >= rhs



from __future__ import annotations

from pydiverse.transform.core import functions
from pydiverse.transform.core.alignment import aligned, eval_aligned
from pydiverse.transform.core.dispatchers import verb
from pydiverse.transform.core.expressions.lambda_getter import λ
from pydiverse.transform.core.table import Table

__all__ = [
    "Table",
    "aligned",
    "eval_aligned",
    "functions",
    "verb",
    "λ",
]



from __future__ import annotations

import sqlalchemy as sa

from pydiverse.transform import ops
from pydiverse.transform._typing import CallableT
from pydiverse.transform.core import dtypes
from pydiverse.transform.core.expressions import TypedValue
from pydiverse.transform.core.registry import TypedOperatorImpl
from pydiverse.transform.core.util import OrderingDescriptor
from pydiverse.transform.errors import OperatorNotSupportedError
from pydiverse.transform.lazy.sql_table.sql_table import SQLTableImpl
from pydiverse.transform.ops import Operator, OPType
from pydiverse.transform.util.warnings import warn_non_standard


class MSSqlTableImpl(SQLTableImpl):
    _dialect_name = "mssql"

    def _build_select_select(self, select):
        s = []
        for name, uuid_ in self.selected_cols():
            sql_col = self.cols[uuid_].compiled(self.sql_columns)
            if not isinstance(sql_col, sa.sql.ColumnElement):
                sql_col = sa.literal(sql_col)
            if dtypes.Bool().same_kind(self.cols[uuid_].dtype):
                # Make sure that any boolean values get stored as bit
                sql_col = sa.cast(sql_col, sa.Boolean())
            s.append(sql_col.label(name))
        return select.with_only_columns(*s)

    def _order_col(
        self, col: sa.SQLColumnExpression, ordering: OrderingDescriptor
    ) -> list[sa.SQLColumnExpression]:
        # MSSQL doesn't support nulls first / nulls last
        order_by_expressions = []

        # asc implies nulls first
        if not ordering.nulls_first and ordering.asc:
            order_by_expressions.append(sa.func.iif(col.is_(None), 1, 0))

        # desc implies nulls last
        if ordering.nulls_first and not ordering.asc:
            order_by_expressions.append(sa.func.iif(col.is_(None), 0, 1))

        order_by_expressions.append(col.asc() if ordering.asc else col.desc())
        return order_by_expressions

    class ExpressionCompiler(SQLTableImpl.ExpressionCompiler):
        def translate(self, expr, **kwargs):
            mssql_bool_as_bit = True
            if verb := kwargs.get("verb"):
                mssql_bool_as_bit = verb not in ("filter", "join")

            return super().translate(
                expr, **kwargs, mssql_bool_as_bit=mssql_bool_as_bit
            )

        def _translate(self, expr, **kwargs):
            if context := kwargs.get("context"):
                if context == "case_val":
                    kwargs["mssql_bool_as_bit"] = True
                elif context == "case_cond":
                    kwargs["mssql_bool_as_bit"] = False

            return super()._translate(expr, **kwargs)

        def _translate_col(self, expr, **kwargs):
            # If mssql_bool_as_bit is true, then we can just return the
            # precompiled col. Otherwise, we must recompile it to ensure
            # we return booleans as bools and not as bits.
            if kwargs.get("mssql_bool_as_bit") is True:
                return super()._translate_col(expr, **kwargs)

            # Can either be a base SQL column, or a reference to an expression
            if expr.uuid in self.backend.sql_columns:
                is_bool = dtypes.Bool().same_kind(self.backend.cols[expr.uuid].dtype)

                def sql_col(cols, **kw):
                    col = cols[expr.uuid]
                    if is_bool:
                        return mssql_convert_bit_to_bool(col)
                    return col

                return TypedValue(sql_col, expr.dtype, OPType.EWISE)

            col = self.backend.cols[expr.uuid]
            return self._translate(col.expr, **kwargs)

        def _translate_function_value(
            self, expr, implementation, op_args, context_kwargs, *, verb=None, **kwargs
        ):
            value = super()._translate_function_value(
                expr,
                implementation,
                op_args,
                context_kwargs,
                verb=verb,
                **kwargs,
            )

            bool_as_bit = kwargs.get("mssql_bool_as_bit")
            returns_bool_as_bit = mssql_op_returns_bool_as_bit(implementation)
            return mssql_convert_bool_bit_value(value, bool_as_bit, returns_bool_as_bit)

        def _translate_function_arguments(self, expr, operator, **kwargs):
            kwargs["mssql_bool_as_bit"] = mssql_op_wants_bool_as_bit(operator)
            return super()._translate_function_arguments(expr, operator, **kwargs)


# Boolean / Bit Conversion
#
# MSSQL doesn't have a boolean type. This means that expressions that
# return a boolean (e.g. ==, !=, >) can't be used in other expressions
# without casting to the BIT type.
# Conversely, after casting to BIT, we sometimes may need to convert
# back to booleans.


def mssql_op_wants_bool_as_bit(operator: Operator) -> bool:
    # These operations want boolean types (not BIT) as input
    exceptions = [
        ops.logical.BooleanBinary,
        ops.logical.Invert,
    ]

    for exception in exceptions:
        if isinstance(operator, exception):
            return False

    return True


def mssql_op_returns_bool_as_bit(implementation: TypedOperatorImpl) -> bool | None:
    if not dtypes.Bool().same_kind(implementation.rtype):
        return None

    # These operations return boolean types (not BIT)
    if isinstance(implementation.operator, ops.logical.Logical):
        return False

    return True


def mssql_convert_bit_to_bool(x: sa.SQLColumnExpression):
    return x == sa.literal_column("1")


def mssql_convert_bool_to_bit(x: sa.SQLColumnExpression):
    return sa.case(
        (x, sa.literal_column("1")),
        (sa.not_(x), sa.literal_column("0")),
    )


def mssql_convert_bool_bit_value(
    value_func: CallableT,
    wants_bool_as_bit: bool | None,
    is_bool_as_bit: bool | None,
) -> CallableT:
    if wants_bool_as_bit is True and is_bool_as_bit is False:

        def value(*args, **kwargs):
            x = value_func(*args, **kwargs)
            return mssql_convert_bool_to_bit(x)

        return value

    if wants_bool_as_bit is False and is_bool_as_bit is True:

        def value(*args, **kwargs):
            x = value_func(*args, **kwargs)
            return mssql_convert_bit_to_bool(x)

        return value

    return value_func


# Operators


with MSSqlTableImpl.op(ops.Equal()) as op:

    @op("str, str -> bool")
    def _eq(x, y):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when comparing strings",
        )
        return x == y


with MSSqlTableImpl.op(ops.NotEqual()) as op:

    @op("str, str -> bool")
    def _ne(x, y):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when comparing strings",
        )
        return x != y


with MSSqlTableImpl.op(ops.Less()) as op:

    @op("str, str -> bool")
    def _lt(x, y):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when comparing strings",
        )
        return x < y


with MSSqlTableImpl.op(ops.LessEqual()) as op:

    @op("str, str -> bool")
    def _le(x, y):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when comparing strings",
        )
        return x <= y


with MSSqlTableImpl.op(ops.Greater()) as op:

    @op("str, str -> bool")
    def _gt(x, y):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when comparing strings",
        )
        return x > y


with MSSqlTableImpl.op(ops.GreaterEqual()) as op:

    @op("str, str -> bool")
    def _ge(x, y):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when comparing strings",
        )
        return x >= y


with MSSqlTableImpl.op(ops.Pow()) as op:

    @op.auto
    def _pow(lhs, rhs):
        # In MSSQL, the output type of pow is the same as the input type.
        # This means, that if lhs is a decimal, then we may very easily loose
        # a lot of precision if the exponent is <= 1
        # https://learn.microsoft.com/en-us/sql/t-sql/functions/power-transact-sql?view=sql-server-ver16
        return sa.func.POWER(sa.cast(lhs, sa.Double()), rhs, type_=sa.Double())


with MSSqlTableImpl.op(ops.RPow()) as op:

    @op.auto
    def _rpow(rhs, lhs):
        return _pow(lhs, rhs)


with MSSqlTableImpl.op(ops.StringLength()) as op:

    @op.auto
    def _str_length(x):
        warn_non_standard(
            "MSSQL ignores trailing whitespace when computing string length",
        )
        return sa.func.LENGTH(x, type_=sa.Integer())


with MSSqlTableImpl.op(ops.Replace()) as op:

    @op.auto
    def _replace(x, y, z):
        x = x.collate("Latin1_General_CS_AS")
        return sa.func.REPLACE(x, y, z, type_=x.type)


with MSSqlTableImpl.op(ops.StartsWith()) as op:

    @op.auto
    def _startswith(x, y):
        x = x.collate("Latin1_General_CS_AS")
        return x.startswith(y, autoescape=True)


with MSSqlTableImpl.op(ops.EndsWith()) as op:

    @op.auto
    def _endswith(x, y):
        x = x.collate("Latin1_General_CS_AS")
        return x.endswith(y, autoescape=True)


with MSSqlTableImpl.op(ops.Contains()) as op:

    @op.auto
    def _contains(x, y):
        x = x.collate("Latin1_General_CS_AS")
        return x.contains(y, autoescape=True)


with MSSqlTableImpl.op(ops.DayOfWeek()) as op:

    @op.auto
    def _day_of_week(x):
        # Offset DOW such that Mon=1, Sun=7
        _1 = sa.literal_column("1")
        _2 = sa.literal_column("2")
        _7 = sa.literal_column("7")
        return (sa.extract("dow", x) + sa.text("@@DATEFIRST") - _2) % _7 + _1


with MSSqlTableImpl.op(ops.Mean()) as op:

    @op.auto
    def _mean(x):
        return sa.func.AVG(sa.cast(x, sa.Double()), type_=sa.Double())


with MSSqlTableImpl.op(ops.StringJoin()) as op:

    @op.auto
    def _join(x, sep: str):
        # We could do something like this:
        #     return sa.func.STRING_AGG(x, sep, type_=x.type).within_group(...)
        # but the problem is, that the StringJoin function is an aggregate function
        # and not a window function, thus we don't (yet) support the `arrange`
        # context kwarg
        raise OperatorNotSupportedError



from __future__ import annotations

import functools
from collections import ChainMap
from collections.abc import Iterable
from typing import Literal

from pydiverse.transform.core import dtypes
from pydiverse.transform.core.dispatchers import builtin_verb
from pydiverse.transform.core.expressions import (
    Column,
    LambdaColumn,
    SymbolicExpression,
)
from pydiverse.transform.core.expressions.util import iterate_over_expr
from pydiverse.transform.core.table_impl import AbstractTableImpl, ColumnMetaData
from pydiverse.transform.core.util import (
    bidict,
    ordered_set,
    sign_peeler,
    translate_ordering,
)
from pydiverse.transform.errors import ExpressionTypeError, FunctionTypeError
from pydiverse.transform.ops import OPType

__all__ = [
    "alias",
    "collect",
    "build_query",
    "show_query",
    "select",
    "rename",
    "mutate",
    "join",
    "left_join",
    "inner_join",
    "outer_join",
    "filter",
    "arrange",
    "group_by",
    "ungroup",
    "summarise",
    "slice_head",
]


def check_cols_available(
    tables: AbstractTableImpl | Iterable[AbstractTableImpl],
    columns: set[Column],
    function_name: str,
):
    if isinstance(tables, AbstractTableImpl):
        tables = (tables,)
    available_columns = ChainMap(*(table.available_cols for table in tables))
    missing_columns = []
    for col in columns:
        if col.uuid not in available_columns:
            missing_columns.append(col)
    if missing_columns:
        missing_columns_str = ", ".join(map(lambda x: str(x), missing_columns))
        raise ValueError(
            f"Can't access column(s) {missing_columns_str} in {function_name}() because"
            " they aren't available in the input."
        )


def check_lambdas_valid(tbl: AbstractTableImpl, *expressions):
    lambdas = []
    for expression in expressions:
        lambdas.extend(
            lc for lc in iterate_over_expr(expression) if isinstance(lc, LambdaColumn)
        )
    missing_lambdas = {lc for lc in lambdas if lc.name not in tbl.named_cols.fwd}
    if missing_lambdas:
        missing_lambdas_str = ", ".join(map(lambda x: str(x), missing_lambdas))
        raise ValueError(f"Invalid lambda column(s) {missing_lambdas_str}.")


def cols_in_expression(expression) -> set[Column]:
    return {c for c in iterate_over_expr(expression) if isinstance(c, Column)}


def cols_in_expressions(expressions) -> set[Column]:
    if len(expressions) == 0:
        return set()
    return set.union(*(cols_in_expression(e) for e in expressions))


def validate_table_args(*tables):
    if len(tables) == 0:
        return

    for table in tables:
        if not isinstance(table, AbstractTableImpl):
            raise TypeError(f"Expected a TableImpl but got {type(table)} instead.")

    backend = type(tables[0])
    for table in tables:
        if type(table) != backend:
            raise ValueError(
                f"Can't mix tables with different backends. Expected '{backend}' but"
                f" found '{type(table)}'."
            )


@builtin_verb()
def alias(tbl: AbstractTableImpl, name: str = None):
    """Creates a new table object with a different name."""
    validate_table_args(tbl)
    return tbl.alias(name)


@builtin_verb()
def collect(tbl: AbstractTableImpl):
    validate_table_args(tbl)
    return tbl.collect()


@builtin_verb()
def build_query(tbl: AbstractTableImpl):
    return tbl.build_query()


@builtin_verb()
def show_query(tbl: AbstractTableImpl):
    if query := tbl.build_query():
        print(query)
    else:
        print(f"No query to show for {type(tbl).__name__}")

    return tbl


@builtin_verb()
def select(tbl: AbstractTableImpl, *args: Column | LambdaColumn):
    if len(args) == 1 and args[0] is Ellipsis:
        # >> select(...)  ->  Select all columns
        args = [
            tbl.cols[uuid].as_column(name, tbl)
            for name, uuid in tbl.named_cols.fwd.items()
        ]

    # Validate input
    validate_table_args(tbl)
    check_cols_available(tbl, cols_in_expressions(args), "select")
    check_lambdas_valid(tbl, *args)

    cols = []
    positive_selection = None
    for col in args:
        col, is_pos = sign_peeler(col)
        if positive_selection is None:
            positive_selection = is_pos
        else:
            if is_pos is not positive_selection:
                raise ValueError(
                    "All columns in input must have the same sign."
                    " Can't mix selection with deselection."
                )

        if not isinstance(col, (Column, LambdaColumn)):
            raise TypeError(
                "Arguments to select verb must be of type 'Column' or 'LambdaColumn'"
                f" and not {type(col)}."
            )
        cols.append(col)

    selects = []
    for col in cols:
        if isinstance(col, Column):
            selects.append(tbl.named_cols.bwd[col.uuid])
        elif isinstance(col, LambdaColumn):
            selects.append(col.name)

    # Invert selection
    if positive_selection is False:
        exclude = set(selects)
        selects.clear()
        for name in tbl.selects:
            if name not in exclude:
                selects.append(name)

    # SELECT
    new_tbl = tbl.copy()
    new_tbl.preverb_hook("select", *args)
    new_tbl.selects = ordered_set(selects)
    new_tbl.select(*args)
    return new_tbl


@builtin_verb()
def rename(tbl: AbstractTableImpl, name_map: dict[str, str]):
    # Type check
    for k, v in name_map.items():
        if not isinstance(k, str) or not isinstance(v, str):
            raise TypeError(
                f"Key and Value of `name_map` must both be strings: ({k!r}, {v!r})"
            )

    # Reference col that doesn't exist
    if missing_cols := name_map.keys() - tbl.named_cols.fwd.keys():
        raise KeyError("Table has no columns named: " + ", ".join(missing_cols))

    # Can't rename two cols to the same name
    _seen = set()
    if duplicate_names := {
        name for name in name_map.values() if name in _seen or _seen.add(name)
    }:
        raise ValueError(
            "Can't rename multiple columns to the same name: "
            + ", ".join(duplicate_names)
        )

    # Can't rename a column to one that already exists
    unmodified_cols = tbl.named_cols.fwd.keys() - name_map.keys()
    if duplicate_names := unmodified_cols & set(name_map.values()):
        raise ValueError(
            "Table already contains columns named: " + ", ".join(duplicate_names)
        )

    # Rename
    new_tbl = tbl.copy()
    new_tbl.selects = ordered_set(name_map.get(name, name) for name in new_tbl.selects)

    uuid_name_map = {new_tbl.named_cols.fwd[old]: new for old, new in name_map.items()}
    for uuid in uuid_name_map:
        del new_tbl.named_cols.bwd[uuid]
    for uuid, name in uuid_name_map.items():
        new_tbl.named_cols.bwd[uuid] = name

    return new_tbl


@builtin_verb()
def mutate(tbl: AbstractTableImpl, **kwargs: SymbolicExpression):
    validate_table_args(tbl)
    check_cols_available(tbl, cols_in_expressions(kwargs.values()), "mutate")

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("mutate", **kwargs)
    kwargs = {k: new_tbl.resolve_lambda_cols(v) for k, v in kwargs.items()}

    for name, expr in kwargs.items():
        uid = Column.generate_col_uuid()
        col = ColumnMetaData.from_expr(uid, expr, new_tbl, verb="mutate")

        if dtypes.NoneDType().same_kind(col.dtype):
            raise ExpressionTypeError(
                f"Column '{name}' has an invalid type: {col.dtype}"
            )

        new_tbl.selects.add(name)
        new_tbl.named_cols.fwd[name] = uid
        new_tbl.available_cols.add(uid)
        new_tbl.cols[uid] = col

    new_tbl.mutate(**kwargs)
    return new_tbl


@builtin_verb()
def join(
    left: AbstractTableImpl,
    right: AbstractTableImpl,
    on: SymbolicExpression,
    how: str,
    *,
    validate: Literal["1:?", None] = None,
):
    validate_table_args(left, right)

    if left.grouped_by or right.grouped_by:
        raise ValueError("Can't join grouped tables. You first have to ungroup them.")

    # Check args only contains valid columns
    check_cols_available((left, right), cols_in_expression(on), "join")

    if how not in ("inner", "left", "outer"):
        raise ValueError(
            "Join type must be one of 'inner', 'left' or 'outer' (value provided:"
            f" {how})"
        )

    # TODO:  Decide on validation options.
    if validate not in (None, "1:?"):
        raise ValueError(
            f"Validate must be either None or '1:?' (value provided: {validate})."
        )

    # Construct new table
    # JOIN -> Merge the left and right table
    new_left = left.copy()
    new_left.preverb_hook("join", right, on, how, validate=validate)

    # Update selects
    right_renamed_selects = ordered_set(
        name + "_" + right.name for name in right.selects
    )
    right_renamed_cols = {
        k + "_" + right.name: v for k, v in right.named_cols.fwd.items()
    }

    # Check for collisions
    # TODO: Review...
    if ambiguous_column_names := new_left.selects & right_renamed_selects:
        raise ValueError("Ambiguous column names: " + ", ".join(ambiguous_column_names))
    if (
        ambiguous_column_names := set(new_left.named_cols.fwd.keys())
        & right_renamed_cols.keys()
    ):
        raise ValueError("Ambiguous column names: " + ", ".join(ambiguous_column_names))
    if ambiguous_column_uuids := set(new_left.named_cols.fwd.values()) & set(
        right_renamed_cols.values()
    ):
        raise ValueError(
            "Ambiguous column uuids: " + ", ".join(map(str, ambiguous_column_uuids))
        )

    new_left.selects |= right_renamed_selects
    new_left.named_cols.fwd.update(right_renamed_cols)
    new_left.available_cols.update(right.available_cols)
    new_left.cols.update(right.cols)

    # By resolving lambdas this late, we enable the user to use lambda columns
    # to reference mutated columns from the right side of the join.
    # -> `λ.columnname_righttablename` is a valid lambda in the on condition.
    check_lambdas_valid(new_left, on)
    on = new_left.resolve_lambda_cols(on)

    new_left.join(right, on, how, validate=validate)
    return new_left


inner_join = functools.partial(join, how="inner")
left_join = functools.partial(join, how="left")
outer_join = functools.partial(join, how="outer")


@builtin_verb()
def filter(tbl: AbstractTableImpl, *args: SymbolicExpression):
    # TODO: Type check expression
    validate_table_args(tbl)
    check_cols_available(tbl, cols_in_expressions(args), "filter")
    args = [tbl.resolve_lambda_cols(arg) for arg in args]

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("filter", *args)
    new_tbl.filter(*args)
    return new_tbl


@builtin_verb()
def arrange(tbl: AbstractTableImpl, *args: Column | LambdaColumn):
    if len(args) == 0:
        return tbl

    # Validate Input
    validate_table_args(tbl)
    check_cols_available(tbl, cols_in_expressions(args), "arrange")
    check_lambdas_valid(tbl, *args)

    ordering = translate_ordering(tbl, args)

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("arrange", *args)
    new_tbl.arrange(ordering)
    return new_tbl


@builtin_verb()
def group_by(tbl: AbstractTableImpl, *args: Column | LambdaColumn, add=False):
    # Validate Input
    validate_table_args(tbl)
    check_cols_available(tbl, cols_in_expressions(args), "group_by")
    check_lambdas_valid(tbl, *args)

    # WARNING: Depending on the SQL backend, you might
    #          only be allowed to reference columns
    if not args:
        raise ValueError(
            "Expected columns to group by, but none were specified. To remove the"
            " grouping use the ungroup verb instead."
        )
    for col in args:
        if not isinstance(col, (Column, LambdaColumn)):
            raise TypeError(
                "Arguments to group_by verb must be of type 'Column' or 'LambdaColumn'"
                f" and not '{type(col)}'."
            )

    args = [tbl.resolve_lambda_cols(arg) for arg in args]

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("group_by", *args, add=add)
    if add:
        new_tbl.grouped_by |= ordered_set(args)
    else:
        new_tbl.grouped_by = ordered_set(args)
    new_tbl.group_by(*args)
    return new_tbl


@builtin_verb()
def ungroup(tbl: AbstractTableImpl):
    """Remove all groupings from table."""
    validate_table_args(tbl)

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("ungroup")
    new_tbl.grouped_by.clear()
    new_tbl.ungroup()
    return new_tbl


@builtin_verb()
def summarise(tbl: AbstractTableImpl, **kwargs: SymbolicExpression):
    # Validate Input
    validate_table_args(tbl)
    check_cols_available(tbl, cols_in_expressions(kwargs.values()), "summarise")

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("summarise", **kwargs)
    kwargs = {k: new_tbl.resolve_lambda_cols(v) for k, v in kwargs.items()}

    # TODO: Validate that the functions are actually aggregating functions.
    ...

    # Calculate state for new table
    selects = ordered_set()
    named_cols = bidict()
    available_cols = set()
    cols = {}

    # Add grouping cols to beginning of select.
    for col in tbl.grouped_by:
        selects.add(tbl.named_cols.bwd[col.uuid])
        available_cols.add(col.uuid)
        named_cols.fwd[col.name] = col.uuid

    # Add summarizing cols to the end of the select.
    for name, expr in kwargs.items():
        if name in selects:
            raise ValueError(
                f"Column with name '{name}' already in select. The new summarised"
                " columns must have a different name than the grouping columns."
            )
        uid = Column.generate_col_uuid()
        col = ColumnMetaData.from_expr(uid, expr, new_tbl, verb="summarise")

        if dtypes.NoneDType().same_kind(col.dtype):
            raise ExpressionTypeError(
                f"Column '{name}' has an invalid type: {col.dtype}"
            )
        if col.ftype != OPType.AGGREGATE:
            raise FunctionTypeError(
                f"Expression for column '{name}' doesn't summarise any values."
            )

        selects.add(name)
        named_cols.fwd[name] = uid
        available_cols.add(uid)
        cols[uid] = col

    # Update new_tbl
    new_tbl.selects = ordered_set(selects)
    new_tbl.named_cols = named_cols
    new_tbl.available_cols = available_cols
    new_tbl.cols.update(cols)
    new_tbl.intrinsic_grouped_by = new_tbl.grouped_by.copy()
    new_tbl.summarise(**kwargs)

    # Reduce the grouping level by one -> drop last
    if len(new_tbl.grouped_by):
        new_tbl.grouped_by.pop_back()

    if len(new_tbl.grouped_by):
        new_tbl.group_by(*new_tbl.grouped_by)
    else:
        new_tbl.ungroup()

    return new_tbl


@builtin_verb()
def slice_head(tbl: AbstractTableImpl, n: int, *, offset: int = 0):
    validate_table_args(tbl)
    if not isinstance(n, int):
        raise TypeError("'n' must be an int")
    if not isinstance(offset, int):
        raise TypeError("'offset' must be an int")
    if n <= 0:
        raise ValueError(f"'n' must be a positive integer (value: {n})")
    if offset < 0:
        raise ValueError(f"'offset' can't be negative (value: {offset})")

    if tbl.grouped_by:
        raise ValueError("Can't slice table that is grouped. Must ungroup first.")

    new_tbl = tbl.copy()
    new_tbl.preverb_hook("slice_head")
    new_tbl.slice_head(n, offset)
    return new_tbl



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

from pydiverse.transform.core.expressions import (
    Column,
    LiteralColumn,
    SymbolicExpression,
    util,
)
from pydiverse.transform.errors import AlignmentError

if TYPE_CHECKING:
    from pydiverse.transform.core import AbstractTableImpl, Table


def aligned(*, with_: str):
    """Decorator for aligned functions."""
    from pydiverse.transform.core import AbstractTableImpl, Table

    if callable(with_):
        raise ValueError("Decorator @aligned requires with_ argument.")

    def decorator(func):
        signature = inspect.signature(func)
        if not isinstance(with_, str):
            raise TypeError(
                f"Argument 'with_' must be of type str, not '{type(with_).__name__}'."
            )
        if with_ not in signature.parameters:
            raise ValueError(f"Function has no argument named '{with_}'")

        def wrapper(*args, **kwargs):
            # Execute func
            result = func(*args, **kwargs)
            if not isinstance(result, SymbolicExpression):
                raise TypeError(
                    "Aligned function must return a symbolic expression not"
                    f" '{result}'."
                )

            # Extract the correct `with_` argument for eval_aligned
            bound_sig = signature.bind(*args, **kwargs)
            bound_sig.apply_defaults()

            alignment_param = bound_sig.arguments[with_]
            if isinstance(alignment_param, SymbolicExpression):
                alignment_param = alignment_param._

            if isinstance(alignment_param, Column):
                aligned_with = alignment_param.table
            elif isinstance(alignment_param, (Table, AbstractTableImpl)):
                aligned_with = alignment_param
            else:
                raise NotImplementedError

            # Evaluate aligned
            return eval_aligned(result, with_=aligned_with)

        return wrapper

    return decorator


def eval_aligned(
    sexpr: SymbolicExpression, with_: AbstractTableImpl | Table = None, **kwargs
) -> SymbolicExpression[LiteralColumn]:
    """Evaluates an expression using the AlignedExpressionEvaluator."""
    from pydiverse.transform.core import AbstractTableImpl, Table

    expr = sexpr._ if isinstance(sexpr, SymbolicExpression) else sexpr

    # Determine Backend
    backend = util.determine_expr_backend(expr)
    if backend is None:
        # TODO: Handle this case. Should return some value...
        raise NotImplementedError

    # Evaluate the function calls on the shared backend
    alignedEvaluator = backend.AlignedExpressionEvaluator(backend.operator_registry)
    result = alignedEvaluator.translate(expr, **kwargs)

    literal_column = LiteralColumn(typed_value=result, expr=expr, backend=backend)

    # Check if alignment condition holds
    if with_ is not None:
        if isinstance(with_, Table):
            with_ = with_._impl
        if not isinstance(with_, AbstractTableImpl):
            raise TypeError(
                "'with_' must either be an instance of a Table or TableImpl. Not"
                f" '{with_}'."
            )

        if not with_.is_aligned_with(literal_column):
            raise AlignmentError(f"Result of eval_aligned isn't aligned with {with_}.")

    # Convert to sexpr so that the user can easily continue transforming
    # it symbolically.
    return SymbolicExpression(literal_column)



from __future__ import annotations

from html import escape
from typing import Any, Generic

from pydiverse.transform._typing import T
from pydiverse.transform.core.expressions import CaseExpression, FunctionCall, util
from pydiverse.transform.core.registry import OperatorRegistry
from pydiverse.transform.core.util import traverse


class SymbolicExpression(Generic[T]):
    """
    Base class to represent a symbolic expression. It can be manipulated using
    standard python operators (for example you can add them) or by calling
    attributes of it.

    To get the non-symbolic version of this expression you use the
    underscore `_` attribute.
    """

    __slots__ = ("_",)

    def __init__(self, underlying: T):
        self._ = underlying

    def __getattr__(self, item) -> SymbolAttribute:
        if item.startswith("_") and item.endswith("_") and len(item) >= 3:
            # Attribute names can't begin and end with an underscore because
            # IPython calls hasattr() to select the correct pretty printing
            # function. Instead of hard coding a specific list, just throw
            # an exception for all attributes that match the general pattern.
            raise AttributeError(
                f"Invalid attribute {item}. Attributes can't begin and end with an"
                " underscore."
            )
        return SymbolAttribute(item, self)

    def __getitem__(self, item):
        return SymbolicExpression(FunctionCall("__getitem__", self, item))

    def case(self, *cases: tuple[Any, Any], default: Any = None):
        case_expression = CaseExpression(
            switching_on=self,
            cases=cases,
            default=default,
        )

        return SymbolicExpression(case_expression)

    def __dir__(self):
        # TODO: Instead of displaying all available operators, translate the
        #       expression and according to the dtype and backend only display
        #       the operators that actually are available.
        return sorted(OperatorRegistry.ALL_REGISTERED_OPS)

    # __contains__, __iter__ and __bool__ are all invalid on s-expressions
    __contains__ = None
    __iter__ = None

    def __bool__(self):
        raise TypeError(
            "Symbolic expressions can't be converted to True/False, "
            "or used with these keywords: not, and, or."
        )

    def __str__(self):
        from pydiverse.transform.core.alignment import eval_aligned

        try:
            result = eval_aligned(self._, check_alignment=False)._

            dtype = result.typed_value.dtype
            value = result.typed_value.value
            return (
                f"Symbolic Expression: {repr(self._)}\ndtype: {dtype}\n\n{str(value)}"
            )
        except Exception as e:
            return (
                f"Symbolic Expression: {repr(self._)}\n"
                "Failed to get evaluate due to an exception:\n"
                f"{type(e).__name__}: {str(e)}"
            )

    def __repr__(self):
        return f"<Sym: {self._}>"

    def _repr_html_(self):
        from pydiverse.transform.core.alignment import eval_aligned

        html = f"<pre>Symbolic Expression:\n{escape(repr(self._))}</pre>"

        try:
            result = eval_aligned(self._, check_alignment=False)._
            backend = util.determine_expr_backend(self._)

            value_repr = backend._html_repr_expr(result.typed_value.value)
            html += (
                f"dtype: <code>{escape(str(result.typed_value.dtype))}</code></br></br>"
            )
            html += f"<pre>{escape(value_repr)}</pre>"
        except Exception as e:
            html += (
                "</br><pre>Failed to get evaluate due to an exception:\n"
                f"{escape(e.__class__.__name__)}: {escape(str(e))}</pre>"
            )

        return html

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else "...")


class SymbolAttribute:
    def __init__(self, name: str, on: SymbolicExpression):
        self.__name = name
        self.__on = on

    def __getattr__(self, item) -> SymbolAttribute:
        return SymbolAttribute(self.__name + "." + item, self.__on)

    def __call__(self, *args, **kwargs) -> SymbolicExpression:
        return SymbolicExpression(FunctionCall(self.__name, self.__on, *args, **kwargs))

    def __hash__(self):
        raise Exception(
            "Nope... You probably didn't want to do this. Did you misspell the"
            f" attribute name '{self.__name}' of '{self.__on}'? Maybe you forgot a"
            " leading underscore."
        )


def unwrap_symbolic_expressions(arg: Any = None):
    """
    Replaces all symbolic expressions in the input with their underlying value.
    """
    return traverse(arg, lambda x: x._ if isinstance(x, SymbolicExpression) else x)


# Add all supported dunder methods to `SymbolicExpression`.
# This has to be done, because Python doesn't call __getattr__ for
# dunder methods.
def create_operator(op):
    def impl(*args, **kwargs):
        return SymbolicExpression(FunctionCall(op, *args, **kwargs))

    return impl


for dunder in OperatorRegistry.SUPPORTED_DUNDER:
    setattr(SymbolicExpression, dunder, create_operator(dunder))
del create_operator



from __future__ import annotations

from pydiverse.transform.ops.core import Binary, ElementWise, Unary

__all__ = [
    "Add",
    "RAdd",
    "Sub",
    "RSub",
    "Mul",
    "RMul",
    "TrueDiv",
    "RTrueDiv",
    "FloorDiv",
    "RFloorDiv",
    "Pow",
    "RPow",
    "Mod",
    "RMod",
    "Neg",
    "Pos",
    "Abs",
    "Round",
]


class Add(ElementWise, Binary):
    name = "__add__"
    signatures = [
        "int, int -> int",
        "int, float -> float",
        "float, int -> float",
        "float, float -> float",
    ]


class RAdd(Add):
    name = "__radd__"


class Sub(ElementWise, Binary):
    name = "__sub__"
    signatures = [
        "int, int -> int",
        "int, float -> float",
        "float, int -> float",
        "float, float -> float",
    ]


class RSub(Sub):
    name = "__rsub__"


class Mul(ElementWise, Binary):
    name = "__mul__"
    signatures = [
        "int, int -> int",
        "int, float -> float",
        "float, int -> float",
        "float, float -> float",
    ]


class RMul(Mul):
    name = "__rmul__"


class TrueDiv(ElementWise, Binary):
    name = "__truediv__"
    signatures = [
        "int, int -> float",
        "float, float -> float",
    ]


class RTrueDiv(TrueDiv):
    name = "__rtruediv__"


class FloorDiv(ElementWise, Binary):
    name = "__floordiv__"
    signatures = [
        "int, int -> int",
    ]


class RFloorDiv(FloorDiv):
    name = "__rfloordiv__"


class Pow(ElementWise, Binary):
    name = "__pow__"
    signatures = [
        "int, int -> float",
        "float, float -> float",
    ]


class RPow(Pow):
    name = "__rpow__"


class Mod(ElementWise, Binary):
    name = "__mod__"
    signatures = [
        "int, int -> int",
    ]


class RMod(Mod):
    name = "__rmod__"


class Neg(ElementWise, Unary):
    name = "__neg__"
    signatures = [
        "int -> int",
        "float -> float",
    ]


class Pos(ElementWise, Unary):
    name = "__pos__"
    signatures = [
        "int -> int",
        "float -> float",
    ]


class Abs(ElementWise, Unary):
    name = "__abs__"
    signatures = [
        "int -> int",
        "float -> float",
    ]


class Round(ElementWise):
    name = "__round__"
    signatures = [
        "int -> int",
        "int, const int -> int",
        "float -> float",
        "float, const int -> float",
    ]



from __future__ import annotations

from pydiverse.transform.core import dtypes
from pydiverse.transform.ops.core import Binary, ElementWise, Operator, Unary

__all__ = [
    "Equal",
    "NotEqual",
    "Less",
    "LessEqual",
    "Greater",
    "GreaterEqual",
    "IsIn",
    "And",
    "RAnd",
    "Or",
    "ROr",
    "Xor",
    "RXor",
    "Invert",
]


class Logical(Operator):
    # Operator that returns a "REAL" boolean.
    # This is mostly relevant for mssql
    pass


#### Comparison Operators ####


class Comparison(ElementWise, Binary, Logical):
    signatures = [
        "int, int -> bool",
        "float, float -> bool",
        "str, str -> bool",
        "bool, bool -> bool",
        "datetime, datetime -> bool",
    ]

    def validate_signature(self, signature):
        assert isinstance(signature.rtype, dtypes.Bool)
        super().validate_signature(signature)


class Equal(Comparison):
    name = "__eq__"
    signatures = Comparison.signatures + [
        "T, const none -> bool",
    ]


class NotEqual(Comparison):
    name = "__ne__"
    signatures = Comparison.signatures + [
        "T, const none -> bool",
    ]


class Less(Comparison):
    name = "__lt__"


class LessEqual(Comparison):
    name = "__le__"


class Greater(Comparison):
    name = "__gt__"


class GreaterEqual(Comparison):
    name = "__ge__"


class IsIn(ElementWise, Logical):
    name = "isin"
    signatures = [
        # TODO: A signature like "T, const list[const T] -> bool" would be better
        "T, const T... -> bool",
    ]


#### Boolean Operators ####


class BooleanBinary(ElementWise, Binary, Logical):
    signatures = [
        "bool, bool -> bool",
    ]

    def validate_signature(self, signature):
        assert len(signature.args) == 2

        for arg_dtype in signature.args:
            assert isinstance(arg_dtype, dtypes.Bool)
            assert not arg_dtype.vararg
            assert not arg_dtype.const

        assert isinstance(signature.rtype, dtypes.Bool)
        super().validate_signature(signature)


class And(BooleanBinary):
    name = "__and__"


class RAnd(BooleanBinary):
    name = "__rand__"


class Or(BooleanBinary):
    name = "__or__"


class ROr(BooleanBinary):
    name = "__ror__"


class Xor(BooleanBinary):
    name = "__xor__"


class RXor(BooleanBinary):
    name = "__rxor__"


class Invert(ElementWise, Unary, Logical):
    name = "__invert__"
    signatures = [
        "bool -> bool",
    ]



from __future__ import annotations

from pydiverse.transform.ops.core import Nullary, Unary, Window

__all__ = [
    "Shift",
    "RowNumber",
    "Rank",
    "DenseRank",
]


class WindowImplicitArrange(Window):
    """
    Like a window function, except that the expression on which this op
    gets called, is used for arranging.

    Converts a call like this ``tbl.col1.nulls_first().rank()``, into a call like this
    ``(tbl.col1).rank(arrange=[tbl.col1.nulls_first()])
    """

    def mutate_args(self, args, kwargs):
        if len(args) == 0:
            return args, kwargs

        from pydiverse.transform.core.util.util import ordering_peeler

        arrange = args[0]
        peeled_first_arg, _, _ = ordering_peeler(arrange)
        args = (peeled_first_arg, *args[1:])

        assert "arrange" not in kwargs
        kwargs = {**kwargs, "arrange": [arrange]}
        return args, kwargs


class Shift(Window):
    name = "shift"
    signatures = [
        "T, const int -> T",
        "T, const int, const T -> T",
    ]


class RowNumber(Window, Nullary):
    name = "row_number"
    signatures = [
        "-> int",
    ]


class Rank(WindowImplicitArrange, Unary):
    name = "rank"
    signatures = [
        "T -> int",
    ]


class DenseRank(WindowImplicitArrange, Unary):
    name = "dense_rank"
    signatures = [
        "T -> int",
    ]



from __future__ import annotations

from pydiverse.transform.ops.core import ElementWise, Unary

__all__ = [
    "Year",
    "Month",
    "Day",
    "Hour",
    "Minute",
    "Second",
    "Millisecond",
    "DayOfWeek",
    "DayOfYear",
]


class DatetimeExtractComponent(ElementWise, Unary):
    signatures = ["datetime -> int"]


class Year(DatetimeExtractComponent):
    name = "year"


class Month(DatetimeExtractComponent):
    name = "month"


class Day(DatetimeExtractComponent):
    name = "day"


class Hour(DatetimeExtractComponent):
    name = "hour"


class Minute(DatetimeExtractComponent):
    name = "minute"


class Second(DatetimeExtractComponent):
    name = "second"


class Millisecond(DatetimeExtractComponent):
    name = "millisecond"


class DayOfWeek(DatetimeExtractComponent):
    name = "day_of_week"


class DayOfYear(DatetimeExtractComponent):
    name = "day_of_year"



from __future__ import annotations

from pydiverse.transform.ops.core import Marker

__all__ = [
    "NullsFirst",
    "NullsLast",
]


# Mark order-by column that it should be ordered with NULLs first
class NullsFirst(Marker):
    name = "nulls_first"
    signatures = ["T -> T"]


# Mark order-by column that it should be ordered with NULLs last
class NullsLast(Marker):
    name = "nulls_last"
    signatures = ["T -> T"]



from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic

from pydiverse.transform._typing import T
from pydiverse.transform.core import registry
from pydiverse.transform.core.expressions import (
    CaseExpression,
    Column,
    FunctionCall,
    LiteralColumn,
)
from pydiverse.transform.ops.core import Operator, OPType
from pydiverse.transform.util import reraise

if TYPE_CHECKING:
    from pydiverse.transform.core.dtypes import DType


# Basic container to store value and associated type metadata
@dataclass
class TypedValue(Generic[T]):
    value: T
    dtype: DType
    ftype: OPType = dataclasses.field(default=OPType.EWISE)

    def __iter__(self):
        return iter((self.value, self.dtype))


class Translator(Generic[T]):
    def translate(self, expr, **kwargs) -> T:
        """Translate an expression recursively."""
        try:
            return bottom_up_replace(expr, lambda e: self._translate(e, **kwargs))
        except Exception as e:
            msg = f"This exception occurred while translating the expression: {expr}"
            reraise(e, suffix=msg)

    def _translate(self, expr, **kwargs) -> T:
        """Translate an expression non recursively."""
        raise NotImplementedError


class DelegatingTranslator(Translator[T], Generic[T]):
    """
    Translator that dispatches to different translate functions based on
    the type of the expression.
    """

    def __init__(self, operator_registry: registry.OperatorRegistry):
        self.operator_registry = operator_registry

    def translate(self, expr, **kwargs):
        """Translate an expression recursively."""
        try:
            return self._translate(expr, **kwargs)
        except Exception as e:
            msg = f"This exception occurred while translating the expression: {expr}"
            reraise(e, suffix=msg)

    def _translate(self, expr, **kwargs):
        if isinstance(expr, Column):
            return self._translate_col(expr, **kwargs)

        if isinstance(expr, LiteralColumn):
            return self._translate_literal_col(expr, **kwargs)

        if isinstance(expr, FunctionCall):
            operator = self.operator_registry.get_operator(expr.name)

            # Mutate function call arguments using operator
            expr_args, expr_kwargs = operator.mutate_args(expr.args, expr.kwargs)
            expr = FunctionCall(expr.name, *expr_args, **expr_kwargs)

            op_args, op_kwargs, context_kwargs = self._translate_function_arguments(
                expr, operator, **kwargs
            )

            if op_kwargs:
                raise NotImplementedError

            signature = tuple(arg.dtype for arg in op_args)
            implementation = self.operator_registry.get_implementation(
                expr.name, signature
            )

            return self._translate_function(
                expr, implementation, op_args, context_kwargs, **kwargs
            )

        if isinstance(expr, CaseExpression):
            switching_on = (
                self._translate(expr.switching_on, **{**kwargs, "context": "case_val"})
                if expr.switching_on is not None
                else None
            )

            cases = []
            for cond, value in expr.cases:
                cases.append(
                    (
                        self._translate(cond, **{**kwargs, "context": "case_cond"}),
                        self._translate(value, **{**kwargs, "context": "case_val"}),
                    )
                )

            default = self._translate(expr.default, **{**kwargs, "context": "case_val"})
            return self._translate_case(expr, switching_on, cases, default, **kwargs)

        if literal_result := self._translate_literal(expr, **kwargs):
            return literal_result

        raise NotImplementedError(
            f"Couldn't find a way to translate object of type {type(expr)} with value"
            f" {expr}."
        )

    def _translate_col(self, expr: Column, **kwargs) -> T:
        raise NotImplementedError

    def _translate_literal_col(self, expr: LiteralColumn, **kwargs) -> T:
        raise NotImplementedError

    def _translate_function(
        self,
        expr: FunctionCall,
        implementation: registry.TypedOperatorImpl,
        op_args: list[T],
        context_kwargs: dict[str, Any],
        **kwargs,
    ) -> T:
        raise NotImplementedError

    def _translate_case(
        self,
        expr: CaseExpression,
        switching_on: T | None,
        cases: list[tuple[T, T]],
        default: T,
        **kwargs,
    ) -> T:
        raise NotImplementedError

    def _translate_literal(self, expr, **kwargs) -> T:
        raise NotImplementedError

    def _translate_function_arguments(
        self, expr: FunctionCall, operator: Operator, **kwargs
    ) -> tuple[list[T], dict[str, T], dict[str, Any]]:
        op_args = [self._translate(arg, **kwargs) for arg in expr.args]
        op_kwargs = {}
        context_kwargs = {}

        for k, v in expr.kwargs.items():
            if k in operator.context_kwargs:
                context_kwargs[k] = v
            else:
                op_kwargs[k] = self._translate(v, **kwargs)

        return op_args, op_kwargs, context_kwargs


def bottom_up_replace(expr, replace):
    def transform(expr):
        if isinstance(expr, FunctionCall):
            f = FunctionCall(
                expr.name,
                *(transform(arg) for arg in expr.args),
                **{k: transform(v) for k, v in expr.kwargs.items()},
            )
            return replace(f)

        if isinstance(expr, CaseExpression):
            c = CaseExpression(
                switching_on=transform(expr.switching_on),
                cases=[(transform(k), transform(v)) for k, v in expr.cases],
                default=transform(expr.default),
            )
            return replace(c)

        return replace(expr)

    return transform(expr)



from __future__ import annotations

import uuid
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Generic

from pydiverse.transform._typing import ImplT, T
from pydiverse.transform.core.dtypes import DType

if TYPE_CHECKING:
    from pydiverse.transform.core.expressions.translator import TypedValue
    from pydiverse.transform.core.table_impl import AbstractTableImpl


def expr_repr(it: Any):
    from pydiverse.transform.core.expressions import SymbolicExpression

    if isinstance(it, SymbolicExpression):
        return expr_repr(it._)
    if isinstance(it, BaseExpression):
        return it._expr_repr()
    if isinstance(it, (list, tuple)):
        return f"[{ ', '.join([expr_repr(e) for e in it]) }]"
    return repr(it)


_dunder_expr_repr = {
    "__add__": lambda lhs, rhs: f"({lhs} + {rhs})",
    "__radd__": lambda rhs, lhs: f"({lhs} + {rhs})",
    "__sub__": lambda lhs, rhs: f"({lhs} - {rhs})",
    "__rsub__": lambda rhs, lhs: f"({lhs} - {rhs})",
    "__mul__": lambda lhs, rhs: f"({lhs} * {rhs})",
    "__rmul__": lambda rhs, lhs: f"({lhs} * {rhs})",
    "__truediv__": lambda lhs, rhs: f"({lhs} / {rhs})",
    "__rtruediv__": lambda rhs, lhs: f"({lhs} / {rhs})",
    "__floordiv__": lambda lhs, rhs: f"({lhs} // {rhs})",
    "__rfloordiv__": lambda rhs, lhs: f"({lhs} // {rhs})",
    "__pow__": lambda lhs, rhs: f"({lhs} ** {rhs})",
    "__rpow__": lambda rhs, lhs: f"({lhs} ** {rhs})",
    "__mod__": lambda lhs, rhs: f"({lhs} % {rhs})",
    "__rmod__": lambda rhs, lhs: f"({lhs} % {rhs})",
    "__round__": lambda x, y=None: f"round({x}, {y})" if y else f"round({x})",
    "__pos__": lambda x: f"(+{x})",
    "__neg__": lambda x: f"(-{x})",
    "__abs__": lambda x: f"abs({x})",
    "__and__": lambda lhs, rhs: f"({lhs} & {rhs})",
    "__rand__": lambda rhs, lhs: f"({lhs} & {rhs})",
    "__or__": lambda lhs, rhs: f"({lhs} | {rhs})",
    "__ror__": lambda rhs, lhs: f"({lhs} | {rhs})",
    "__xor__": lambda lhs, rhs: f"({lhs} ^ {rhs})",
    "__rxor__": lambda rhs, lhs: f"({lhs} ^ {rhs})",
    "__invert__": lambda x: f"(~{x})",
    "__lt__": lambda lhs, rhs: f"({lhs} < {rhs})",
    "__le__": lambda lhs, rhs: f"({lhs} <= {rhs})",
    "__eq__": lambda lhs, rhs: f"({lhs} == {rhs})",
    "__ne__": lambda lhs, rhs: f"({lhs} != {rhs})",
    "__gt__": lambda lhs, rhs: f"({lhs} > {rhs})",
    "__ge__": lambda lhs, rhs: f"({lhs} >= {rhs})",
}


class BaseExpression:
    def _expr_repr(self) -> str:
        """String repr that, when executed, returns the same expression"""
        raise NotImplementedError


class Column(BaseExpression, Generic[ImplT]):
    __slots__ = ("name", "table", "dtype", "uuid")

    def __init__(self, name: str, table: ImplT, dtype: DType, uuid: uuid.UUID = None):
        self.name = name
        self.table = table
        self.dtype = dtype
        self.uuid = uuid or Column.generate_col_uuid()

    def __repr__(self):
        return f"<{self.table.name}.{self.name}({self.dtype})>"

    def _expr_repr(self) -> str:
        return f"{self.table.name}.{self.name}"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.name == other.name and self.uuid == other.uuid

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.uuid)

    @classmethod
    def generate_col_uuid(cls) -> uuid.UUID:
        return uuid.uuid1()


class LambdaColumn(BaseExpression):
    """Anonymous Column

    A lambda column is a column without an associated table or UUID. This means
    that it can be used to reference columns in the same pipe as it was created.

    Example:
      The following fails because `table.a` gets referenced before it gets created.
        table >> mutate(a = table.x) >> mutate(b = table.a)
      Instead you can use a lambda column to achieve this:
        table >> mutate(a = table.x) >> mutate(b = λ.a)
    """

    __slots__ = "name"

    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return f"<λ.{self.name}>"

    def _expr_repr(self) -> str:
        return f"λ.{self.name}"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(("λ", self.name))


class LiteralColumn(BaseExpression, Generic[T]):
    __slots__ = ("typed_value", "expr", "backend")

    def __init__(
        self,
        typed_value: TypedValue[T],
        expr: Any,
        backend: type[AbstractTableImpl],
    ):
        self.typed_value = typed_value
        self.expr = expr
        self.backend = backend

    def __repr__(self):
        return f"<Lit: {self.expr} ({self.typed_value.dtype})>"

    def _expr_repr(self) -> str:
        return repr(self)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.typed_value == other.typed_value
            and self.expr == other.expr
            and self.backend == other.backend
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class FunctionCall(BaseExpression):
    """
    AST node to represent a function / operator call.
    """

    def __init__(self, name: str, *args, **kwargs):
        from pydiverse.transform.core.expressions.symbolic_expressions import (
            unwrap_symbolic_expressions,
        )

        # Unwrap all symbolic expressions in the input
        args = unwrap_symbolic_expressions(args)
        kwargs = unwrap_symbolic_expressions(kwargs)

        self.name = name
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        args = [repr(e) for e in self.args] + [
            f"{k}={repr(v)}" for k, v in self.kwargs.items()
        ]
        return f'{self.name}({", ".join(args)})'

    def _expr_repr(self) -> str:
        args = [expr_repr(e) for e in self.args] + [
            f"{k}={expr_repr(v)}" for k, v in self.kwargs.items()
        ]

        if self.name in _dunder_expr_repr:
            return _dunder_expr_repr[self.name](*args)

        if len(self.args) == 0:
            args_str = ", ".join(args)
            return f"f.{self.name}({args_str})"
        else:
            args_str = ", ".join(args[1:])
            return f"{args[0]}.{self.name}({args_str})"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.name, self.args, tuple(self.kwargs.items())))

    def iter_children(self):
        yield from self.args


class CaseExpression(BaseExpression):
    def __init__(
        self, switching_on: Any | None, cases: Iterable[tuple[Any, Any]], default: Any
    ):
        from pydiverse.transform.core.expressions.symbolic_expressions import (
            unwrap_symbolic_expressions,
        )

        # Unwrap all symbolic expressions in the input
        switching_on = unwrap_symbolic_expressions(switching_on)
        cases = unwrap_symbolic_expressions(list(cases))
        default = unwrap_symbolic_expressions(default)

        self.switching_on = switching_on
        self.cases = cases
        self.default = default

    def __repr__(self):
        if self.switching_on:
            return f"case({self.switching_on}, {self.cases}, default={self.default})"
        else:
            return f"case({self.cases}, default={self.default})"

    def _expr_repr(self) -> str:
        prefix = "f"
        if self.switching_on:
            prefix = expr_repr(self.switching_on)

        args = [expr_repr(case) for case in self.cases]
        args.append(f"default={expr_repr(self.default)}")
        return f"{prefix}.case({', '.join(args)})"

    def iter_children(self):
        if self.switching_on:
            yield self.switching_on

        for k, v in self.cases:
            yield k
            yield v

        yield self.default



from __future__ import annotations

import enum
from collections import ChainMap
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pydiverse.transform.core.registry import OperatorSignature

__all__ = [
    "OPType",
    "Operator",
    "OperatorExtension",
    "Arity",
    "Nullary",
    "Unary",
    "Binary",
    "ElementWise",
    "Aggregate",
    "Window",
    "Marker",
]


class OPType(enum.IntEnum):
    EWISE = 1
    AGGREGATE = 2
    WINDOW = 3


_OPERATOR_VALID = "_operator_valid_"
_OPERATOR_MISSING_ATTR = "_operator_missing_attr_"


class Operator:
    """
    Base class to define an operator. All class level attributes that have the
    value `NotImplemented` must be set or else the operator can't be initialized.

    name:
        The name of the operator. The main mechanism to access an operator is
        using `(s-expression).operator_name(args)`.

    ftype:
        The operator type (function type). Can either be an element wise,
        aggregate or window function.

    signatures:
        A list of default signatures for this operator.

    context_kwargs:
        A set of keyword argument names. Can later be used by the translator
        to handle these keyword arguments specially.

    """

    name: str = NotImplemented
    ftype: OPType = NotImplemented
    signatures: list[str] = None
    context_kwargs: set[str] = None

    def __new__(cls, *args, **kwargs):
        if not getattr(cls, _OPERATOR_VALID, False):
            missing = getattr(cls, _OPERATOR_MISSING_ATTR, None)
            raise NotImplementedError(
                "Can't initialize operator. Some class attributes are not implemented:"
                f" {missing}."
            )
        return super().__new__(cls, *args, **kwargs)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        mro_dict = ChainMap(*(sup.__dict__ for sup in cls.__mro__))
        not_implemented = [k for k, v in mro_dict.items() if v is NotImplemented]
        setattr(cls, _OPERATOR_VALID, len(not_implemented) == 0)
        setattr(cls, _OPERATOR_MISSING_ATTR, not_implemented)

    def __eq__(self, other):
        return type(self) == type(other)

    def __hash__(self):
        return hash(type(self))

    def validate_signature(self, signature: OperatorSignature):
        pass

    def mutate_args(self, args, kwargs):
        """
        Allows the operator to modify the arguments passed to it before translation
        """
        return args, kwargs


class OperatorExtension:
    """
    An extension to an operator. Provides additional signatures.
    """

    operator: type[Operator] = NotImplemented
    signatures: list[str] = NotImplemented

    def __new__(cls, *args, **kwargs):
        if cls.operator == NotImplemented or cls.signatures == NotImplemented:
            raise NotImplementedError(
                "Can't initialize operator extension. "
                "Either the operator or signatures attribute is missing."
            )
        if not issubclass(cls.operator, Operator):
            raise TypeError
        return super().__new__(cls, *args, **kwargs)


# Arity


class Arity(Operator):
    """Base class for checking the number of arguments."""

    n_arguments: int = NotImplemented

    def validate_signature(self, signature):
        assert len(signature.args) == self.n_arguments
        super().validate_signature(signature)


class Nullary(Arity):
    n_arguments = 0


class Unary(Arity):
    n_arguments = 1


class Binary(Arity):
    n_arguments = 2


# Base operator types


class ElementWise(Operator):
    ftype = OPType.EWISE


class Aggregate(Operator):
    ftype = OPType.AGGREGATE


class Window(Operator):
    ftype = OPType.WINDOW
    context_kwargs = {
        "arrange",  # List[Column | LambdaColumn]
    }


class Marker(Operator):
    ftype = None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic

from pydiverse.transform._typing import ImplT
from pydiverse.transform.core.table_impl import AbstractTableImpl


@dataclass
class JoinDescriptor(Generic[ImplT]):
    __slots__ = ("right", "on", "how")

    right: ImplT
    on: Any
    how: str


class LazyTableImpl(AbstractTableImpl):
    """Base class for lazy backends."""

    pass



from __future__ import annotations

from pydiverse.transform.core.expressions import LambdaColumn
from pydiverse.transform.core.expressions.symbolic_expressions import SymbolicExpression


class LambdaColumnGetter:
    """
    An instance of this object can be used to instantiate a LambdaColumn.
    """

    def __getattr__(self, item):
        if item.startswith("__"):
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{item}'"
            )
        return SymbolicExpression(LambdaColumn(item))

    def __getitem__(self, item):
        return SymbolicExpression(LambdaColumn(item))


# Global instance of LambdaColumnGetter.
λ = LambdaColumnGetter()



from __future__ import annotations

from typing import TYPE_CHECKING

from pydiverse.transform.core.expressions import (
    CaseExpression,
    Column,
    FunctionCall,
    LiteralColumn,
)

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from pydiverse.transform.core.table_impl import AbstractTableImpl


def iterate_over_expr(expr, expand_literal_col=False):
    """
    Iterate in depth-first preorder over the expression and yield all components.
    """

    yield expr

    if isinstance(expr, FunctionCall):
        for child in expr.iter_children():
            yield from iterate_over_expr(child, expand_literal_col=expand_literal_col)
        return

    if isinstance(expr, CaseExpression):
        for child in expr.iter_children():
            yield from iterate_over_expr(child, expand_literal_col=expand_literal_col)
        return

    if expand_literal_col and isinstance(expr, LiteralColumn):
        yield from iterate_over_expr(expr.expr, expand_literal_col=expand_literal_col)
        return


def determine_expr_backend(expr) -> type[AbstractTableImpl] | None:
    """Returns the backend used in an expression.

    Iterates over an expression and extracts the underlying backend type used.
    If no backend can be determined (because the expression doesn't contain a
    column), None is returned instead. If different backends are being used,
    throws an exception.
    """

    backends = set()
    for atom in iterate_over_expr(expr):
        if isinstance(atom, Column):
            backends.add(type(atom.table))
        if isinstance(atom, LiteralColumn):
            backends.add(atom.backend)

    if len(backends) == 1:
        return backends.pop()
    if len(backends) >= 2:
        raise ValueError(
            "Expression contains different backends "
            f"(found: {[backend.__name__ for backend in backends]})."
        )
    return None



from __future__ import annotations

from pydiverse.transform.ops.core import Aggregate, Unary

__all__ = [
    "Min",
    "Max",
    "Mean",
    "Sum",
    "Any",
    "All",
    "StringJoin",
    "Count",
]


class Min(Aggregate, Unary):
    name = "min"
    signatures = [
        "int -> int",
        "float -> float",
        "str -> str",
        "datetime -> datetime",
    ]


class Max(Aggregate, Unary):
    name = "max"
    signatures = [
        "int -> int",
        "float -> float",
        "str -> str",
        "datetime -> datetime",
    ]


class Mean(Aggregate, Unary):
    name = "mean"
    signatures = [
        "int -> float",
        "float -> float",
    ]


class Sum(Aggregate, Unary):
    name = "sum"
    signatures = [
        "int -> int",
        "float -> float",
    ]


class Any(Aggregate, Unary):
    name = "any"
    signatures = [
        "bool -> bool",
    ]


class All(Aggregate, Unary):
    name = "all"
    signatures = [
        "bool -> bool",
    ]


class StringJoin(Aggregate):
    name = "join"
    signatures = [
        "str, const str -> str",
    ]


class Count(Aggregate):
    name = "count"
    signatures = [
        "-> int",
        "T -> int",
    ]



from __future__ import annotations


class OperatorNotSupportedError(Exception):
    """
    Exception raised when a specific operation is not supported by a backend.
    """


class ExpressionError(Exception):
    """
    Generic exception related to an invalid expression.
    """


class ExpressionTypeError(ExpressionError):
    """
    Exception related to invalid types in an expression
    """


class FunctionTypeError(ExpressionError):
    """
    Exception related to function type
    """


class AlignmentError(Exception):
    """
    Raised when something isn't aligned.
    """


# WARNINGS


class NonStandardBehaviourWarning(UserWarning):
    """
    Category for when a specific backend deviates from
    the expected standard behaviour.
    """



from __future__ import annotations

from typing import Any

from pydiverse.transform.core.expressions import (
    CaseExpression,
    FunctionCall,
    SymbolicExpression,
)

__all__ = [
    "count",
    "row_number",
]


def _sym_f_call(name, *args, **kwargs) -> SymbolicExpression[FunctionCall]:
    return SymbolicExpression(FunctionCall(name, *args, **kwargs))


def count(expr: SymbolicExpression = None):
    if expr is None:
        return _sym_f_call("count")
    else:
        return _sym_f_call("count", expr)


def row_number(*, arrange: list):
    return _sym_f_call("row_number", arrange=arrange)


def case(*cases: tuple[Any, Any], default: Any = None):
    case_expression = CaseExpression(
        switching_on=None,
        cases=cases,
        default=default,
    )

    return SymbolicExpression(case_expression)


def min(first: Any, *expr: Any):
    return _sym_f_call("__least", first, *expr)


def max(first: Any, *expr: Any):
    return _sym_f_call("__greatest", first, *expr)



from __future__ import annotations

from pydiverse.transform.core.table_impl import AbstractTableImpl


def uuid_to_str(_uuid):
    # mod with 2^31-1  (prime number)
    return format(_uuid.int % 0x7FFFFFFF, "X")


class EagerTableImpl(AbstractTableImpl):
    pass



from __future__ import annotations

from collections.abc import Iterable
from html import escape
from typing import Generic

from pydiverse.transform._typing import ImplT
from pydiverse.transform.core.expressions import (
    Column,
    LambdaColumn,
    SymbolicExpression,
)


class Table(Generic[ImplT]):
    """
    All attributes of a table are columns except for the `_impl` attribute
    which is a reference to the underlying table implementation.
    """

    def __init__(self, implementation: ImplT):
        self._impl = implementation

    def __getitem__(self, key) -> SymbolicExpression[Column]:
        if isinstance(key, SymbolicExpression):
            key = key._
        return SymbolicExpression(self._impl.get_col(key))

    def __setitem__(self, col, expr):
        """Mutate a column
        :param col: Either a str or SymbolicColumn
        """
        from pydiverse.transform.core.verbs import mutate

        col_name = None

        if isinstance(col, SymbolicExpression):
            underlying = col._
            if isinstance(underlying, (Column, LambdaColumn)):
                col_name = underlying.name
        elif isinstance(col, str):
            col_name = col

        if not col_name:
            raise KeyError(
                f"Invalid key {col}. Must be either a string, Column or LambdaColumn."
            )
        self._impl = (self >> mutate(**{col_name: expr}))._impl

    def __getattr__(self, name) -> SymbolicExpression[Column]:
        return SymbolicExpression(self._impl.get_col(name))

    def __iter__(self) -> Iterable[SymbolicExpression[Column]]:
        # Capture current state (this allows modifying the table inside a loop)
        cols = [
            SymbolicExpression(self._impl.get_col(name))
            for name, _ in self._impl.selected_cols()
        ]
        return iter(cols)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self._impl == other._impl

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._impl)

    def __dir__(self):
        return sorted(self._impl.named_cols.fwd.keys())

    def __contains__(self, item):
        if isinstance(item, SymbolicExpression):
            item = item._
        if isinstance(item, LambdaColumn):
            return item.name in self._impl.named_cols.fwd
        if isinstance(item, Column):
            return item.uuid in self._impl.available_cols
        return False

    def __copy__(self):
        impl_copy = self._impl.copy()
        return self.__class__(impl_copy)

    def __str__(self):
        from pydiverse.transform.core.verbs import collect

        try:
            return (
                f"Table: {self._impl.name}, backend: {type(self._impl).__name__}\n"
                f"{self >> collect()}"
            )
        except Exception as e:
            return (
                f"Table: {self._impl.name}, backend: {type(self._impl).__name__}\n"
                "Failed to collect table due to an exception:\n"
                f"{type(e).__name__}: {str(e)}"
            )

    def _repr_html_(self) -> str | None:
        from pydiverse.transform.core.verbs import collect

        html = (
            f"Table <code>{self._impl.name}</code> using"
            f" <code>{type(self._impl).__name__}</code> backend:</br>"
        )
        try:
            # TODO: For lazy backend only show preview (eg. take first 20 rows)
            html += (self >> collect())._repr_html_()
        except Exception as e:
            html += (
                "</br><pre>Failed to collect table due to an exception:\n"
                f"{escape(e.__class__.__name__)}: {escape(str(e))}</pre>"
            )
        return html

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else "...")



from __future__ import annotations

from collections.abc import Iterable, MutableSet

from pydiverse.transform._typing import T


class ordered_set(MutableSet[T]):
    def __init__(self, values: Iterable[T] = tuple()):
        self.__data = {v: None for v in values}

    def __contains__(self, item: T) -> bool:
        return item in self.__data

    def __iter__(self) -> Iterable[T]:
        yield from self.__data.keys()

    def __len__(self) -> int:
        return len(self.__data)

    def __repr__(self):
        return "{%s}" % ", ".join(repr(e) for e in self)

    def __copy__(self):
        return self.__class__(self)

    def add(self, value: T) -> None:
        self.__data[value] = None

    def discard(self, value: T) -> None:
        del self.__data[value]

    def clear(self) -> None:
        self.__data.clear()

    def copy(self):
        return self.__copy__()

    def pop_back(self) -> None:
        """Return the popped value.Raise KeyError if empty."""
        if len(self) == 0:
            raise KeyError("Ordered set is empty.")
        back = next(reversed(self.__data.keys()))
        self.discard(back)
        return back



from __future__ import annotations

import typing
from dataclasses import dataclass

from pydiverse.transform._typing import T
from pydiverse.transform.core.expressions import FunctionCall

__all__ = (
    "traverse",
    "sign_peeler",
    "OrderingDescriptor",
    "translate_ordering",
)


def traverse(obj: T, callback: typing.Callable) -> T:
    if isinstance(obj, list):
        return [traverse(elem, callback) for elem in obj]
    if isinstance(obj, dict):
        return {k: traverse(v, callback) for k, v in obj.items()}
    if isinstance(obj, tuple):
        if type(obj) != tuple:
            # Named tuples cause problems
            raise Exception
        return tuple(traverse(elem, callback) for elem in obj)

    return callback(obj)


def peel_markers(expr, markers):
    found_markers = []
    while isinstance(expr, FunctionCall):
        if expr.name in markers:
            found_markers.append(expr.name)
            assert len(expr.args) == 1
            expr = expr.args[0]
        else:
            break
    return expr, found_markers


def sign_peeler(expr):
    """
    Remove unary - and + prefix and return the sign
    :return: `True` for `+` and `False` for `-`
    """

    expr, markers = peel_markers(expr, {"__neg__", "__pos__"})
    num_neg = markers.count("__neg__")
    return expr, num_neg % 2 == 0


def ordering_peeler(expr):
    expr, markers = peel_markers(
        expr, {"__neg__", "__pos__", "nulls_first", "nulls_last"}
    )

    ascending = markers.count("__neg__") % 2 == 0
    nulls_first = False
    for marker in markers:
        if marker == "nulls_first":
            nulls_first = True
            break
        if marker == "nulls_last":
            break

    return expr, ascending, nulls_first


####


@dataclass
class OrderingDescriptor:
    __slots__ = ("order", "asc", "nulls_first")

    order: typing.Any
    asc: bool
    nulls_first: bool


def translate_ordering(tbl, order_list) -> list[OrderingDescriptor]:
    ordering = []
    for arg in order_list:
        col, ascending, nulls_first = ordering_peeler(arg)
        col = tbl.resolve_lambda_cols(col)
        ordering.append(OrderingDescriptor(col, ascending, nulls_first))

    return ordering



from __future__ import annotations

from pydiverse.transform.ops.core import ElementWise

__all__ = [
    "Greatest",
    "Least",
]


class Greatest(ElementWise):
    name = "__greatest"
    signatures = ["T... -> T"]


class Least(ElementWise):
    name = "__least"
    signatures = ["T... -> T"]



from __future__ import annotations

import sqlalchemy as sa

from pydiverse.transform import ops
from pydiverse.transform.lazy.sql_table.sql_table import SQLTableImpl
from pydiverse.transform.util.warnings import warn_non_standard


class SQLiteTableImpl(SQLTableImpl):
    _dialect_name = "sqlite"


with SQLiteTableImpl.op(ops.Round()) as op:

    @op.auto
    def _round(x, decimals=0):
        if decimals >= 0:
            return sa.func.ROUND(x, decimals, type_=x.type)
        # For some reason SQLite doesn't like negative decimals values
        return sa.func.ROUND(x / (10**-decimals), type_=x.type) * (10**-decimals)


with SQLiteTableImpl.op(ops.StartsWith()) as op:

    @op.auto
    def _startswith(x, y):
        warn_non_standard(
            "SQLite: startswith is case-insensitive by default. "
            "Use the 'case_sensitive_like' pragma to change this behaviour. "
            "See https://www.sqlite.org/pragma.html#pragma_case_sensitive_like",
        )
        return x.startswith(y, autoescape=True)


with SQLiteTableImpl.op(ops.EndsWith()) as op:

    @op.auto
    def _endswith(x, y):
        warn_non_standard(
            "SQLite: endswith is case-insensitive by default. "
            "Use the 'case_sensitive_like' pragma to change this behaviour. "
            "See https://www.sqlite.org/pragma.html#pragma_case_sensitive_like",
        )
        return x.endswith(y, autoescape=True)


with SQLiteTableImpl.op(ops.Contains()) as op:

    @op.auto
    def _contains(x, y):
        warn_non_standard(
            "SQLite: contains is case-insensitive by default. "
            "Use the 'case_sensitive_like' pragma to change this behaviour. "
            "See https://www.sqlite.org/pragma.html#pragma_case_sensitive_like",
        )
        return x.contains(y, autoescape=True)


with SQLiteTableImpl.op(ops.Millisecond()) as op:

    @op.auto
    def _millisecond(x):
        warn_non_standard(
            "SQLite returns rounded milliseconds",
        )
        _1000 = sa.literal_column("1000")
        frac_seconds = sa.cast(sa.func.STRFTIME("%f", x), sa.Numeric())
        return sa.cast((frac_seconds * _1000) % _1000, sa.Integer())


with SQLiteTableImpl.op(ops.Greatest()) as op:

    @op.auto
    def _greatest(*x):
        # The SQLite MAX function returns NULL if any of the inputs are NULL
        # -> Use divide and conquer approach with coalesce to ensure correct result
        if len(x) == 1:
            return x[0]

        mid = (len(x) + 1) // 2
        left = _greatest(*x[:mid])
        right = _greatest(*x[mid:])

        # TODO: Determine return type
        return sa.func.coalesce(sa.func.MAX(left, right), left, right)


with SQLiteTableImpl.op(ops.Least()) as op:

    @op.auto
    def _least(*x):
        # The SQLite MIN function returns NULL if any of the inputs are NULL
        # -> Use divide and conquer approach with coalesce to ensure correct result
        if len(x) == 1:
            return x[0]

        mid = (len(x) + 1) // 2
        left = _least(*x[:mid])
        right = _least(*x[mid:])

        # TODO: Determine return type
        return sa.func.coalesce(sa.func.MIN(left, right), left, right)


with SQLiteTableImpl.op(ops.StringJoin()) as op:

    @op.auto
    def _join(x, sep: str):
        return sa.func.GROUP_CONCAT(x, sep, type_=x.type)



from __future__ import annotations

from pydiverse.transform.ops.core import ElementWise, OperatorExtension, Unary
from pydiverse.transform.ops.logical import Logical
from pydiverse.transform.ops.numeric import Add, RAdd

__all__ = [
    "StringAdd",
    "StringRAdd",
    "Strip",
    "StringLength",
    "Upper",
    "Lower",
    "Replace",
    "StartsWith",
    "EndsWith",
    "Contains",
]


class StringAdd(OperatorExtension):
    operator = Add
    signatures = [
        "str, str -> str",
    ]


class StringRAdd(OperatorExtension):
    operator = RAdd
    signatures = [
        "str, str -> str",
    ]


####


class StringUnary(ElementWise, Unary):
    signatures = [
        "str -> str",
    ]


class Strip(StringUnary):
    name = "strip"


class StringLength(StringUnary):
    name = "len"
    signatures = [
        "str -> int",
    ]


class Upper(StringUnary):
    name = "upper"


class Lower(StringUnary):
    name = "lower"


class Replace(ElementWise):
    name = "replace"
    signatures = [
        "str, const str, const str -> str",
    ]


class StartsWith(ElementWise, Logical):
    name = "startswith"
    signatures = [
        "str, const str -> bool",
    ]


class EndsWith(ElementWise, Logical):
    name = "endswith"
    signatures = [
        "str, const str -> bool",
    ]


class Contains(ElementWise, Logical):
    name = "contains"
    signatures = [
        "str, const str -> bool",
    ]
